__all__ = ('ScheduledEvent', )

from ...bases import DiscordEntity, ICON_TYPE_NONE, IconSlot
from ...channel import ChannelType, create_partial_channel_from_id
from ...core import GUILDS, SCHEDULED_EVENTS
from ...http import urls as module_urls
from ...precreate_helpers import process_precreate_parameters
from ...user import ZEROUSER

from ..scheduled_event_entity_metadata import ScheduledEventEntityMetadataBase

from .fields import (
    parse_channel_id, parse_creator, parse_description, parse_end, parse_entity_id, parse_entity_metadata,
    parse_entity_type, parse_guild_id, parse_id, parse_name, parse_privacy_level, parse_sku_ids, parse_start,
    parse_status, parse_user_count, put_channel_id_into, put_creator_into, put_description_into, put_end_into,
    put_entity_id_into, put_entity_metadata_into, put_entity_type_into, put_guild_id_into, put_id_into, put_name_into,
    put_privacy_level_into, put_sku_ids_into, put_start_into, put_status_into, put_user_count_into, validate_channel_id,
    validate_creator, validate_description, validate_end, validate_entity_id, validate_entity_type, validate_guild_id,
    validate_id, validate_name, validate_privacy_level, validate_sku_ids, validate_start, validate_status,
    validate_user_count
)
from .helpers import guess_scheduled_event_entity_type_from_keyword_parameters
from .preinstanced import PrivacyLevel, ScheduledEventEntityType, ScheduledEventStatus


SCHEDULED_EVENT_IMAGE = IconSlot(
    'image',
    'image',
    module_urls.scheduled_event_image_url,
    module_urls.scheduled_event_image_url_as,
)


PRECREATE_FIELDS = {
    'channel': ('channel_id', validate_channel_id),
    'channel_id': ('channel_id', validate_channel_id),
    'creator': ('creator', validate_creator),
    'description': ('description', validate_description),
    'end': ('end', validate_end),
    'entity_id': ('entity_id', validate_entity_id),
    'guild': ('guild_id', validate_guild_id),
    'guild_id': ('guild_id', validate_guild_id),
    'image': ('image', SCHEDULED_EVENT_IMAGE.validate_icon),
    'name': ('name', validate_name),
    'privacy_level': ('privacy_level', validate_privacy_level),
    'start': ('start', validate_start),
    'sku_ids': ('sku_ids', validate_sku_ids),
    'status': ('status', validate_status),
    'user_count': ('user_count', validate_user_count),
}


class ScheduledEvent(DiscordEntity):
    """
    Attributes
    ----------
    channel_id : `int`
        The event's stage's channel identifier.
    creator : ``ClientUserBase``
        The event's creator.
    description : `None`, `str`
        Description of the event.
    end : `None`, `datetime`
        The scheduled end time of the event.
    entity_id : `int`
        The event's entity's identifier.
    entity_metadata : ``ScheduledEventEntityMetadataBase``
        Metadata about the target entity.
    entity_type : ``ScheduledEventEntityType``
        To which type of entity the event is bound to.
    guild_id : `int`
        The respective event's identifier.
    id : `int`
        The scheduled event's identifier number.
    image_type : ``IconType``
        The event's image's type.
    image_hash : `int`
        The event's image hash.
    name : `str`
        The event's name.
    privacy_level : ``PrivacyLevel``
        The privacy level of the event.
    start : `None`, `datetime`
        The scheduled start time of the event.
    sku_ids : `None`, `tuple` of `int`
        Stock keeping unit identifiers used at the event.
    status : ``ScheduledEventStatus``
        The status of the event.
    user_count : `int`
        Users subscribed to the event.
    
    Notes
    -----
    Scheduled event instances support weakreferencing.
    """
    __slots__ = (
        '__weakref__', 'channel_id', 'creator', 'description', 'end', 'entity_id', 'entity_metadata', 'entity_type',
        'guild_id', 'name', 'privacy_level', 'sku_ids', 'start', 'status', 'user_count'
    )
    
    image = SCHEDULED_EVENT_IMAGE
    
    def __new__(
        cls,
        *,
        channel_id = ...,
        description = ...,
        end = ...,
        entity_type = ...,
        image = ...,
        name = ...,
        privacy_level = ...,
        start = ...,
        status = ...,
        **keyword_parameters,
    ):
        """
        Creates a new partial scheduled event with the given fields.
        
        Parameters
        ----------
        channel_id : `int`, ``Channel``, Optional (Keyword only)
            The event's stage's channel or its identifier.
        description : `None`, `str`, Optional (Keyword only)
            Description of the event.
        end : `None`, `datetime`, Optional (Keyword only)
            The scheduled end time of the event.
        entity_type : ``ScheduledEventEntityType``, `int`, Optional (Keyword only)
            To which type of entity the event is bound to.
        image : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The schedule event's image.
        name : `str`
            The event's name.
        privacy_level : ``PrivacyLevel``, `int`, Optional (Keyword only)
            The privacy level of the event.
        start : `None`, `datetime`, Optional (Keyword only)
            The scheduled start time of the event.
        status : ``ScheduledEventStatus``, Optional (Keyword only)
            The status of the event.
        **keyword_parameters : Keyword parameters
            Additional keyword parameters passed to the entity metadata.
        
        Other Parameters
        ----------------
        location : `None`, `str`, Optional (Keyword only)
            The place where the event will take place.
        speaker_ids : `None`, `iterable` of (`int`, `ClientUserBase`), Optional (Keyword only)
            The speakers' identifier of the stage channel.
        
        Raises
        ------
        TypeError
            - Extra or unused parameters.
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # channel_id
        if channel_id is ...:
            channel_id = 0
        else:
            channel_id = validate_channel_id(channel_id)
        
        # description
        if description is ...:
            description = None
        else:
            description = validate_description(description)
        
        # end
        if end is ...:
            end = None
        else:
            end = validate_end(end)
        
        # entity_type
        if entity_type is ...:
            entity_type = guess_scheduled_event_entity_type_from_keyword_parameters(keyword_parameters)
        else:
            entity_type = validate_entity_type(entity_type)
        
        # image
        if image is ...:
            image = None
        else:
            image = cls.image.validate_icon(image, allow_data = True)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # privacy_level
        if privacy_level is ...:
            privacy_level = PrivacyLevel.guild_only
        else:
            privacy_level = validate_privacy_level(privacy_level)
        
        # start
        if start is ...:
            start = None
        else:
            start = validate_start(start)
        
        # status
        if status is ...:
            status = ScheduledEventStatus.none
        else:
            status = validate_status(status)
        
        # entity_metadata
        entity_metadata = entity_type.metadata_type.from_keyword_parameters(keyword_parameters)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused keyword parameters: {keyword_parameters!r}.'
            )
        
        # Construct
        self = object.__new__(cls)
        self.channel_id = channel_id
        self.creator = ZEROUSER
        self.description = description
        self.end = end
        self.entity_id = 0
        self.entity_metadata = entity_metadata
        self.entity_type = entity_type
        self.guild_id = 0
        self.id = 0
        self.image = image
        self.name = name
        self.privacy_level = privacy_level
        self.start = start
        self.sku_ids = None
        self.status = status
        self.user_count = 0
        return self
    
    
    @classmethod
    def from_data(cls, data, *, strong_cache = True):
        """
        Creates a new scheduled event instance from the received data.
        
        If the instance already exists, returns that instead.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Guild scheduled event data.
        strong_cache : `bool` = `True`, Optional (Keyword only)
            Whether the instance should be put into its strong cache.
        
        Returns
        -------
        self : `instance<cls>`
        """
        scheduled_event_id = parse_id(data)
        
        try:
            self = SCHEDULED_EVENTS[scheduled_event_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = scheduled_event_id
            self._set_attributes(data)
            SCHEDULED_EVENTS[scheduled_event_id] = self
        else:
            if strong_cache and (not self.partial):
                self._update_counts_only(data)
                return self
        
            self._set_attributes(data)
        
        if strong_cache:
            try:
                guild = GUILDS[self.guild_id]
            except KeyError:
                pass
            else:
                guild.scheduled_events[scheduled_event_id] = self
        
        return self
    
    
    @classmethod
    def from_data_is_created(cls, data):
        """
        Creates a new scheduled event instance from the received data. If already exists, picks it up.
        
        Also returns whether the instance was new (or partial) or already existed.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Guild scheduled event data.
        
        Returns
        -------
        self : `instance<cls>`
        is_created : `bool`
        """
        scheduled_event_id = parse_id(data)
        
        try:
            self = SCHEDULED_EVENTS[scheduled_event_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = scheduled_event_id
            SCHEDULED_EVENTS[scheduled_event_id] = self
        
        else:
            if not self.partial:
                self._update_counts_only(data)
                return self, False
        
        self._set_attributes(data)
        
        try:
            guild = GUILDS[self.guild_id]
        except KeyError:
            pass
        else:
            guild.scheduled_events[scheduled_event_id] = self
        
        return self, True
    
    
    @classmethod
    def _create_from_data_and_delete(cls, data):
        """
        Creates a new scheduled event instance from the received data.
        
        If the instance already exists, returns that instead.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Guild event data.
        
        Returns
        -------
        self : `instance<cls>`
            The created or found scheduled event instance.
        """
        scheduled_event_id = parse_id(data)
        
        try:
            self = SCHEDULED_EVENTS[scheduled_event_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = scheduled_event_id
            SCHEDULED_EVENTS[scheduled_event_id] = self
        else:
            if self._delete():
                self._update_counts_only(data)
                return self
        
        self._set_attributes(data)
        return self
    
    
    def _delete(self):
        """
        Tries to delete the scheduled event from its guild.
        
        Returns
        -------
        deleted : `bool`
        """
        try:
            guild = GUILDS[self.guild_id]
        except KeyError:
            return False
            
        try:
            del guild.scheduled_events[self.id]
        except KeyError:
            return False
        
        return True
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the scheduled event to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        
        put_channel_id_into(self.channel_id, data, defaults)
        put_description_into(self.description, data, defaults)
        put_end_into(self.end, data, defaults)
        put_entity_metadata_into(self.entity_metadata, data, defaults)
        put_entity_type_into(self.entity_type, data, defaults)
        type(self).image.put_into(self.image, data, defaults, as_data = not include_internals)
        put_name_into(self.name, data, defaults)
        put_privacy_level_into(self.privacy_level, data, defaults)
        put_start_into(self.start, data, defaults)
        put_status_into(self.status, data, defaults)
        
        if include_internals:
            put_creator_into(self.creator, data, defaults, include_internals = include_internals)
            put_entity_id_into(self.entity_id, data, defaults)
            put_guild_id_into(self.guild_id, data, defaults)
            put_id_into(self.id, data, defaults)
            put_sku_ids_into(self.sku_ids, data, defaults)
            put_user_count_into(self.user_count, data, defaults)
        
        return data
    
    
    def _set_attributes(self, data):
        """
        Sets the event's attributes form the received data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Scheduled event data.
        """
        self.creator = parse_creator(data)
        self.guild_id = parse_guild_id(data)
        self.user_count = parse_user_count(data)
        
        self._update_attributes(data)
    
    
    def _update_attributes(self, data):
        """
        Updates the attributes of the scheduled event.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Scheduled event data.
        """
        self._set_image(data)
        
        self.channel_id = parse_channel_id(data)
        self.description = parse_description(data)
        self.end = parse_end(data)
        self.entity_id = parse_entity_id(data)
        self.name = parse_name(data)
        self.privacy_level = parse_privacy_level(data)
        self.sku_ids = parse_sku_ids(data)
        self.start = parse_start(data)
        self.status = parse_status(data)
        
        self.entity_type = entity_type = parse_entity_type(data)
        self.entity_metadata = parse_entity_metadata(data, entity_type)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the attributes of the scheduled event and returns the changed ones within an `attribute-name` -
        `old-value` relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Scheduled event data.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `object`) items
            The updated attributes.
            
            The returned dictionary might contain the following items:
            
            +---------------------------+-----------------------------------------------+
            | Key                       | Value                                         |
            +===========================+===============================================+
            | channel_id                | `int`                                         |
            +---------------------------+-----------------------------------------------+
            | description               | `None`, `str`                                 |
            +---------------------------+-----------------------------------------------+
            | end                       | `None`, `datetime`                            |
            +---------------------------+-----------------------------------------------+
            | entity_id                 | `int`                                         |
            +---------------------------+-----------------------------------------------+
            | entity_metadata           | ``ScheduledEventEntityMetadataBase``          |
            +---------------------------+-----------------------------------------------+
            | entity_type               | ``ScheduledEventEntityType``                  |
            +---------------------------+-----------------------------------------------+
            | image                     | ``Icon``                                      |
            +---------------------------+-----------------------------------------------+
            | name                      | `str`                                         |
            +---------------------------+-----------------------------------------------+
            | privacy_level             | ``PrivacyLevel``                              |
            +---------------------------+-----------------------------------------------+
            | sku_ids                   | `None`, `tuple` of `int`                      |
            +---------------------------+-----------------------------------------------+
            | start                     | `None`, `datetime`                            |
            +---------------------------+-----------------------------------------------+
            | status                    | ``ScheduledEventStatus``                      |
            +---------------------------+-----------------------------------------------+
        """
        old_attributes = {}
        
        self._update_image(data, old_attributes)
        
        channel_id = parse_channel_id(data)
        if self.channel_id != channel_id:
            old_attributes['channel_id'] = self.channel_id
            self.channel_id = channel_id
        
        description = parse_description(data)
        if self.description != description:
            old_attributes['description'] = self.description
            self.description = description
        
        end = parse_end(data)
        if self.end != end:
            old_attributes['end'] = self.end
            self.end = end
        
        entity_id = parse_entity_id(data)
        if self.entity_id != entity_id:
            old_attributes['entity_id'] = self.entity_id
            self.entity_id = entity_id
        
        name = parse_name(data)
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        privacy_level = parse_privacy_level(data)
        if self.privacy_level is not privacy_level:
            old_attributes['privacy_level'] = self.privacy_level
            self.privacy_level = privacy_level
        
        sku_ids = parse_sku_ids(data)
        if self.sku_ids != sku_ids:
            old_attributes['sku_ids'] = self.sku_ids
            self.sku_ids = sku_ids
        
        start = parse_start(data)
        if self.start != start:
            old_attributes['start'] = self.start
            self.start = start
        
        status =  parse_status(data)
        if self.status is not status:
            old_attributes['status'] = self.status
            self.status = status
        
        
        entity_type = parse_entity_type(data)
        if self.entity_type is not entity_type:
            old_attributes['entity_type'] = self.entity_type
            self.entity_type = entity_type
        
        entity_metadata = parse_entity_metadata(data, entity_type)
        if self.entity_metadata != entity_metadata:
            old_attributes['entity_metadata'] = self.entity_metadata
            self.entity_metadata = entity_metadata
        
        return old_attributes
    
    
    def _update_counts_only(self, data):
        """
        Updates the scheduled event's count attributes only.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Scheduled event data.
        """
        try:
            user_count = data['user_count']
        except KeyError:
            pass
        else:
            self.user_count = user_count
        
    
    def __repr__(self):
        """Returns the guild event's representation."""
        repr_parts = [
            '<', self.__class__.__name__,
            ' id = ', repr(self.id),
            ', name = ', repr(self.name),
            ', guild_id = ', repr(self.guild_id),
        ]
        
        entity_type = self.entity_type
        if entity_type is not ScheduledEventEntityType.none:
            repr_parts.append(', entity_type = ')
            repr_parts.append(entity_type.name)
            repr_parts.append(' ~ ')
            repr_parts.append(repr(entity_type.value))
        
        
        status = self.status
        if status is not ScheduledEventStatus.none:
            repr_parts.append(', status = ')
            repr_parts.append(status.name)
            repr_parts.append(' ~ ')
            repr_parts.append(repr(status.value))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the scheduled event's hash value."""
        stage_id = self.id
        if stage_id:
            return stage_id
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Returns the scheduled event's hash value.
        This function is called when the scheduled event is partial.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # channel_id
        hash_value ^= self.channel_id
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= hash(self.description)
        
        # end
        end = self.end
        if (end is not None):
            hash_value ^= hash(self.end)
        
        # entity_metadata
        hash_value ^= hash(self.entity_metadata)
        
        # entity_type
        hash_value ^= hash(self.entity_type)
        
        # image
        hash_value ^ hash(self.image)
        
        # name
        name = self.name
        if (name != description):
            hash_value ^= hash(name)
        
        hash_value ^= hash(self.privacy_level) << 4
        # scheduled_event_id | internal
        
        # start
        start = self.start
        if (start is not None):
            hash_value ^= hash(self.start)
        
        return hash_value
    

    def __eq__(self, other):
        """Returns whether the two scheduled events are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two scheduled events are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether self is equal to other. Other must be same type as self.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        is_equal : `bool`
        """
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            if self.id != other.id:
                return False
        
        # channel_id
        if self.channel_id != other.channel_id:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # end
        if self.end != other.end:
            return False
        
        # entity_type
        if self.entity_type != other.entity_type:
            return False
        
        # entity_metadata
        if self.entity_metadata != other.entity_metadata:
            return False
        
        # image
        if self.image != other.image:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # privacy_level
        if self.privacy_level is not other.privacy_level:
            return False
        
        # start
        if self.start != other.start:
            return False
        
        # status
        if self.status is not other.status:
            return False
        
        return True
        
        
    @classmethod
    def _create_empty(cls, scheduled_event_id):
        """
        Creates a scheduled event instance with its attributes set to their default values.
        
        Parameters
        ----------
        scheduled_event_id : `int`
            The stage's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.channel_id = 0
        self.creator = ZEROUSER
        self.description = None
        self.end = None
        self.entity_id = 0
        self.entity_metadata = ScheduledEventEntityMetadataBase._create_empty()
        self.entity_type = ScheduledEventEntityType.none
        self.guild_id = 0
        self.id = scheduled_event_id
        self.image_hash = 0
        self.image_type = ICON_TYPE_NONE
        self.name = ''
        self.privacy_level = PrivacyLevel.guild_only
        self.start = None
        self.sku_ids = None
        self.status = ScheduledEventStatus.none
        self.user_count = 0
        return self
    
    
    @classmethod
    def precreate(cls, scheduled_event_id, *, entity_type = ..., **keyword_parameters):
        """
        Precreates the scheduled event by creating a partial one with the given parameters.
        When the scheduled event is loaded the precreated one will be picked up. If an already existing scheduled event
        would be precreated, returns that instead and updates that only, if that is partial.
        
        Parameters
        ----------
        scheduled_event_id : `int`
            The scheduled event's identifier.
        entity_type : ``ScheduledEventEntityType``, `int`, Optional (Keyword only)
            To which type of entity the event is bound to.
        **keyword_parameters : keyword parameters
            Additional predefined attributes for the scheduled event.
        
        Other Parameters
        ----------------
        channel_id : `int`, ``Channel``, Optional (Keyword only)
            Alternative for `channel`.
        channel_id : `int`, ``Channel``, Optional (Keyword only)
            The event's stage's channel or its identifier.
        creator : ``ClientUserBase``, Optional (Keyword only)
            The event's creator.
        description : `None`, `str`, Optional (Keyword only)
            Description of the event.
        end : `None`, `datetime`, Optional (Keyword only)
            The scheduled end time of the event.
        entity_id : `int`, Optional (Keyword only)
            The event's entity's identifier.
        entity_type : ``ScheduledEventEntityType``, `int`, Optional (Keyword only)
            To which type of entity the event is bound to.
        guild : `int`, ``Guild``, Optional (Keyword only)
            Alternative for `guild_id`.
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The scheduled event's guild or its identifier.
        image : `None`, ``Icon``, `str`, Optional (Keyword only)
            The schedule event's image.
        location : `None`, `str`, Optional (Keyword only)
            The place where the event will take place.
        name : `str`
            The event's name.
        privacy_level : ``PrivacyLevel``, `int`, Optional (Keyword only)
            The privacy level of the event.
        sku_ids : `None`, `iterable` of `int`, Optional (Keyword only)
            Stock keeping unit identifiers used at the event.
        speaker_ids : `None`, `iterable` of (`int`, `ClientUserBase`), Optional (Keyword only)
            The speakers' identifier of the stage channel.
        start : `None`, `datetime`, Optional (Keyword only)
            The scheduled start time of the event.
        status : ``ScheduledEventStatus``, Optional (Keyword only)
            The status of the event.
        user_count : `int`, Optional (Keyword only)
            Users subscribed to the event.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        scheduled_event_id = validate_id(scheduled_event_id)

        if (entity_type is not ...) or keyword_parameters:
            processed = []
                
            # entity_type
            if entity_type is ...:
                entity_type = guess_scheduled_event_entity_type_from_keyword_parameters(keyword_parameters)
            else:
                entity_type = validate_entity_type(entity_type)
            processed.append(('entity_type', entity_type))
            
            # keyword_parameters
            extra = process_precreate_parameters(keyword_parameters, PRECREATE_FIELDS, processed)
            
            # entity_metadata
            entity_metadata_type = entity_type.metadata_type
            if entity_metadata_type is not ScheduledEventEntityMetadataBase:
                if extra is None:
                    entity_metadata = entity_metadata_type._create_empty()
                else:
                    entity_metadata = entity_metadata_type.from_keyword_parameters(extra)
                processed.append(('entity_metadata', entity_metadata))
            
            # raise on extra
            if (extra is not None) and extra:
                raise TypeError(
                    f'Extra or unused keyword parameters: {keyword_parameters!r}.'
                )
        
        else:
            processed = None
        
        try:
            self = SCHEDULED_EVENTS[scheduled_event_id]
        except KeyError:
            self = cls._create_empty(scheduled_event_id)
            SCHEDULED_EVENTS[scheduled_event_id] = self
        else:
            if not self.partial:
                return self
        
        if (processed is not None):
            for name, value in processed:
                setattr(self, name, value)
        
        return self
    
    
    def copy(self):
        """
        Copies the scheduled event.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.channel_id = self.channel_id
        new.creator = ZEROUSER
        new.description = self.description
        new.end = self.end
        new.entity_id = 0
        new.entity_metadata = self.entity_metadata.copy()
        new.entity_type = self.entity_type
        new.guild_id = 0
        new.id = 0
        new.image = self.image
        new.name = self.name
        new.privacy_level = self.privacy_level
        new.start = self.start
        new.sku_ids = None
        new.status = self.status
        new.user_count = 0
        return new
    
    
    def copy_with(
        self,
        channel_id = ...,
        description = ...,
        end = ...,
        entity_type = ...,
        image = ...,
        name = ...,
        privacy_level = ...,
        start = ...,
        status = ...,
        **keyword_parameters,
    ):
        """
        Copies the scheduled event with the given fields.
        
        Parameters
        ----------
        channel_id : `int`, ``Channel``, Optional (Keyword only)
            The event's stage's channel or its identifier.
        description : `None`, `str`, Optional (Keyword only)
            Description of the event.
        end : `None`, `datetime`, Optional (Keyword only)
            The scheduled end time of the event.
        entity_type : ``ScheduledEventEntityType``, `int`, Optional (Keyword only)
            To which type of entity the event is bound to.
        image : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The schedule event's image.
        name : `str`
            The event's name.
        privacy_level : ``PrivacyLevel``, `int`, Optional (Keyword only)
            The privacy level of the event.
        start : `None`, `datetime`, Optional (Keyword only)
            The scheduled start time of the event.
        status : ``ScheduledEventStatus``, Optional (Keyword only)
            The status of the event.
        **keyword_parameters : Keyword parameters
            Additional keyword parameters passed to the entity metadata.
        
        Other Parameters
        ----------------
        location : `None`, `str`, Optional (Keyword only)
            The place where the event will take place.
        speaker_ids : `None`, `iterable` of (`int`, `ClientUserBase`), Optional (Keyword only)
            The speakers' identifier of the stage channel.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's value is incorrect.
            - Extra or unused parameters.
        ValueError
            - If a parameter's type is incorrect.
        """
        # channel_id
        if channel_id is ...:
            channel_id = self.channel_id
        else:
            channel_id = validate_channel_id(channel_id)
        
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # end
        if end is ...:
            end = self.end
        else:
            end = validate_end(end)
        
        # entity_type
        if entity_type is ...:
            entity_type = guess_scheduled_event_entity_type_from_keyword_parameters(keyword_parameters)
            if (entity_type is ScheduledEventEntityType.none):
                entity_type = self.entity_type
        else:
            entity_type = validate_entity_type(entity_type)
        
        # image
        if image is ...:
            image = self.image
        else:
            image = type(self).image.validate_icon(image, allow_data = True)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # privacy_level
        if privacy_level is ...:
            privacy_level = self.privacy_level
        else:
            privacy_level = validate_privacy_level(privacy_level)
        
        # start
        if start is ...:
            start = self.start
        else:
            start = validate_start(start)
        
        # status
        if status is ...:
            status = self.status
        else:
            status = validate_status(status)
        
        # entity_metadata
        entity_metadata = self.entity_metadata
        if (entity_type.metadata_type is type(entity_metadata)):
            entity_metadata = entity_metadata.copy_with_keyword_parameters(keyword_parameters)
        else:
            entity_metadata = entity_type.metadata_type.from_keyword_parameters(keyword_parameters)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused keyword parameters: {keyword_parameters!r}.'
            )
    
        # Construct
        new = object.__new__(type(self))
        new.channel_id = channel_id
        new.creator = ZEROUSER
        new.description = description
        new.end = end
        new.entity_id = 0
        new.entity_metadata = entity_metadata
        new.entity_type = entity_type
        new.guild_id = 0
        new.id = 0
        new.image = image
        new.name = name
        new.privacy_level = privacy_level
        new.start = start
        new.sku_ids = None
        new.status = status
        new.user_count = 0
        return new
    
    
    url = property(module_urls.scheduled_event_url)
    
    
    @property
    def partial(self):
        """
        Returns whether the scheduled event is partial.
        
        Returns
        -------
        partial : `bool`
        """
        guild_id = self.guild_id
        if guild_id:
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                pass
            else:
                if not guild.partial:
                    if self.id in guild.scheduled_events.keys():
                        return False
        
        return True
    
    
    @property
    def entity(self):
        """
        Returns the stage channel's entity if applicable.
        
        Returns
        -------
        entity : `None`, ``Channel``
        """
        entity_id = self.entity_id
        if entity_id:
            entity_type = self.entity_type
            if entity_type is ScheduledEventEntityType.none:
                entity = None
            elif entity_type is ScheduledEventEntityType.stage:
                entity = create_partial_channel_from_id(entity_id, ChannelType.guild_stage, self.guild_id)
            elif entity_type is ScheduledEventEntityType.voice:
                entity = create_partial_channel_from_id(entity_id, ChannelType.guild_voice, self.guild_id)
            else:
                entity = None
        else:
            entity = None
        
        return entity
    
    
    @property
    def channel(self):
        """
        Returns the event's channel if has any.
        
        Returns
        -------
        channel : `None`, ``Channel``
        """
        channel_id = self.channel_id
        if channel_id:
            return create_partial_channel_from_id(channel_id, ChannelType.unknown, self.guild_id)
    
    
    @property
    def guild(self):
        """
        Returns the event's guild.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        return GUILDS.get(self.guild_id, None)
    
    
    @property
    def creator_id(self):
        """
        The event's creator's identifier.
        
        Returns
        -------
        user_id : `int`
        """
        return self.creator.id
    
    
    def iter_sku_ids(self):
        """
        Iterates over the stock keeping unit identifiers used at the event.
        
        This method is an iterable generator.
        
        Yields
        ------
        sku_id : `int`
        """
        sku_ids = self.sku_ids
        if (sku_ids is not None):
            yield from sku_ids



__all__ = ('unload_all_plugin',)

from scarletio import copy_docs

from ..plugin_loader import PLUGIN_LOADER


@copy_docs(PLUGIN_LOADER.load)
def unload_all_plugin(*, blocking = True):
    return PLUGIN_LOADER.unload_all(blocking = blocking)



__all__ = ('Application', )

import warnings
from functools import partial as partial_func

from scarletio import BaseMethodDescriptor, export

from ...bases import DiscordEntity, ICON_TYPE_NONE, IconSlot
from ...core import APPLICATION_ID_TO_CLIENT, APPLICATIONS
from ...http import urls as module_urls
from ...precreate_helpers import process_precreate_parameters_and_raise_extra
from ...user import ZEROUSER

from .constants import (
    BOT_PUBLIC_DEFAULT, BOT_REQUIRE_CODE_GRANT_DEFAULT, HOOK_DEFAULT, MAX_PARTICIPANTS_DEFAULT,
    OVERLAY_COMPATIBILITY_HOOK_DEFAULT, OVERLAY_DEFAULT
)
from .fields import (
    parse_aliases, parse_bot_public, parse_bot_require_code_grant, parse_custom_install_url, parse_deeplink_url,
    parse_description, parse_developers, parse_eula_id, parse_executables, parse_flags, parse_guild_id, parse_hook,
    parse_id, parse_install_parameters, parse_max_participants, parse_name, parse_overlay,
    parse_overlay_compatibility_hook, parse_owner, parse_primary_sku_id, parse_privacy_policy_url, parse_publishers,
    parse_role_connection_verification_url, parse_rpc_origins, parse_slug, parse_tags, parse_terms_of_service_url,
    parse_third_party_skus, parse_type, parse_verify_key, put_aliases_into, put_bot_public_into,
    put_bot_require_code_grant_into, put_custom_install_url_into, put_deeplink_url_into, put_description_into,
    put_developers_into, put_eula_id_into, put_executables_into, put_flags_into, put_guild_id_into, put_hook_into,
    put_id_into, put_install_parameters_into, put_max_participants_into, put_name_into,
    put_overlay_compatibility_hook_into, put_overlay_into, put_owner_into, put_primary_sku_id_into,
    put_privacy_policy_url_into, put_publishers_into, put_role_connection_verification_url_into, put_rpc_origins_into,
    put_slug_into, put_tags_into, put_terms_of_service_url_into, put_third_party_skus_into, put_type_into,
    put_verify_key_into, validate_aliases, validate_bot_public, validate_bot_require_code_grant,
    validate_custom_install_url, validate_deeplink_url, validate_description, validate_developers, validate_eula_id,
    validate_executables, validate_flags, validate_guild_id, validate_hook, validate_id, validate_install_parameters,
    validate_max_participants, validate_name, validate_overlay, validate_overlay_compatibility_hook, validate_owner,
    validate_primary_sku_id, validate_privacy_policy_url, validate_publishers,
    validate_role_connection_verification_url, validate_rpc_origins, validate_slug, validate_tags,
    validate_terms_of_service_url, validate_third_party_skus, validate_type, validate_verify_key
)
from .flags import ApplicationFlag
from .preinstanced import ApplicationType

# Invite application fields
#
# - bot_public
# - bot_require_code_grant
# - cover_image
# - description
# - flags
# - hook
# - icon
# - id
# - max_participants
# - name
# - privacy_policy_url
# - rpc_origins
# - splash
# - summary (Deprecated)
# - tags
# - type
# - terms_of_service_url
# - verify_key
#
# Own application fields
#
# - bot_public
# - bot_require_code_grant
# - custom_install_url
# - description
# - flags
# - guild_id
# - hook
# - icon
# - id
# - name
# - owner
# - primary_sku_id
# - slug
# - summary (deprecated)
# - tags
# - team
# - type
# - verify_key
#
# Extra from docs
# - role_connection_verification_url
# - rpc_origins
# - terms_of_service_url
# - privacy_policy_url
# - cover_image
# - install_params
#
# Detectable application fields
#
# - aliases
# - bot_public
# - bot_require_code_grant
# - cover_image
# - deeplink_uri
# - description
# - developers
# - eula_id
# - executables
# - flags
# - guild_id
# - hook
# - icon
# - id
# - name
# - overlay
# - overlay_compatibility_hook
# - primary_sku_id
# - privacy_policy_url
# - publishers
# - rpc_origins
# - slug
# - splash
# - summary (deprecated)
# - tags
# - terms_of_service_url
# - third_party_skus
# - type
# - verify_key
#
# Table format:
#
# +-------------------------------------+-----------+-----------+---------------+
# | Name                                | Own       | Invite    | Detectable    |
# +=====================================+===========+===========+===============+
# | aliases                             | NO        | NO        | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | bot_public                          | YES       | YES       | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | bot_require_code_grant              | YES       | YES       | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | cover_image                         | PROBABLY  | YES       | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | custom_install_url                  | YES       | NO        | NO            |
# +-------------------------------------+-----------+-----------+---------------+
# | deeplink_uri                        | NO        | NO        | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | description                         | YES       | YES       | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | developers                          | NO        | NO        | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | eula_id                             | NO        | NO        | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | executables                         | NO        | NO        | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | flags                               | YES       | YES       | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | guild_id                            | YES       | NO        | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | hook                                | YES       | YES       | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | icon                                | YES       | YES       | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | id                                  | YES       | YES       | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | install_params                      | PROBABLY  | NO        | NO            |
# +-------------------------------------+-----------+-----------+---------------+
# | max_participants                    | NO        | YES       | NO            |
# +-------------------------------------+-----------+-----------+---------------+
# | name                                | YES       | YES       | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | overlay                             | NO        | NO        | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | overlay_compatibility_hook          | NO        | NO        | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | owner                               | YES       | NO        | NO            |
# +-------------------------------------+-----------+-----------+---------------+
# | primary_sku_id                      | YES       | NO        | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | privacy_policy_url                  | PROBABLY  | YES       | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | publishers                          | NO        | NO        | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | role_connection_verification_url    | PROBABLY  | NO        | NO            |
# +-------------------------------------+-----------+-----------+---------------+
# | rpc_origins                         | YES       | YES       | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | slug                                | YES       | NO        | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | splash                              | PROBABLY  | YES       | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | summary                             | YES       | YES       | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | tags                                | YES       | YES       | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | team                                | YES       | NO        | NO            |
# +-------------------------------------+-----------+-----------+---------------+
# | terms_of_service_url                | PROBABLY  | YES       | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | third_party_skus                    | NO        | NO        | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | type                                | YES       | YES       | YES           |
# +-------------------------------------+-----------+-----------+---------------+
# | verify_key                          | YES       | YES       | YES           |
# +-------------------------------------+-----------+-----------+---------------+


application_cover = IconSlot(
    'cover',
    'cover_image',
    module_urls.application_cover_url,
    module_urls.application_cover_url_as,
    add_updater = False,
)

application_icon = IconSlot(
    'icon',
    'icon',
    module_urls.application_icon_url,
    module_urls.application_icon_url_as,
    add_updater = False,
)

application_splash = IconSlot(
    'splash',
    'splash',
    None,
    None,
    add_updater = False,
)


COMMON_CONSTRUCT_FIELDS = {
    'aliases': ('aliases', validate_aliases),
    'bot_public': ('bot_public', validate_bot_public),
    'bot_require_code_grant': ('bot_require_code_grant', validate_bot_require_code_grant),
    'custom_install_url': ('custom_install_url', validate_custom_install_url),
    'deeplink_url': ('deeplink_url', validate_deeplink_url),
    'description': ('description', validate_description),
    'developers': ('developers', validate_developers),
    'eula_id': ('eula_id', validate_eula_id),
    'executables': ('executables', validate_executables),
    'flags': ('flags', validate_flags),
    'guild_id': ('guild_id', validate_guild_id),
    'hook': ('hook', validate_hook),
    'install_parameters': ('install_parameters', validate_install_parameters),
    'max_participants': ('max_participants', validate_max_participants),
    'name': ('name', validate_name),
    'overlay': ('overlay', validate_overlay),
    'overlay_compatibility_hook': ('overlay_compatibility_hook', validate_overlay_compatibility_hook),
    'owner': ('owner', validate_owner),
    'primary_sku_id': ('primary_sku_id', validate_primary_sku_id),
    'privacy_policy_url': ('privacy_policy_url', validate_privacy_policy_url),
    'publishers': ('publishers', validate_publishers),
    'role_connection_verification_url': ('role_connection_verification_url', validate_role_connection_verification_url),
    'rpc_origins': ('rpc_origins', validate_rpc_origins),
    'slug': ('slug', validate_slug),
    'tags': ('tags', validate_tags),
    'terms_of_service_url': ('terms_of_service_url', validate_terms_of_service_url),
    'third_party_skus': ('third_party_skus', validate_third_party_skus),
    'application_type': ('type', validate_type),
    'verify_key': ('verify_key', validate_verify_key),
}

PRECREATE_FIELDS = {
    **COMMON_CONSTRUCT_FIELDS,
    'cover': ('cover', application_cover.validate_icon),
    'icon': ('icon', application_icon.validate_icon),
    'splash': ('splash', application_splash.validate_icon),
}

NEW_FIELDS = {
    **COMMON_CONSTRUCT_FIELDS,
    'cover': ('cover', partial_func(application_cover.validate_icon, allow_data = True)),
    'icon': ('icon', partial_func(application_icon.validate_icon, allow_data = True)),
    'splash': ('splash', partial_func(application_splash.validate_icon, allow_data = True)),
}


@export
class Application(DiscordEntity, immortal = True):
    """
    Represents a Discord application with all of it's spice.
    
    When a ``Client`` is created, it starts it's life with an empty application by default. However when the client
    logs in, it's application is requested, but it can be updated by ``Client.update_application_info`` anytime.
    
    Attributes
    ----------
    aliases : `None`, `tuple` of `str`
        Aliases of the application's name.
    
    bot_public : `bool`.
        Whether not only the application's owner can join the application's bot to guilds.
        Defaults to `False`
    
    bot_require_code_grant : `bool`
        Whether the application's bot will only join a guild, when completing the full `oauth2` code grant flow.
        Defaults to `False`.
    
    cover_hash : `int`
        The application's store cover image's hash in `uint128`. If the application is sold at Discord, this image
        will be used at the store.
    
    cover_type : ``IconType``
        The application's store cover image's type.
    
    custom_install_url : `None`, `str`
        The application's default custom authorization link if enabled.
        Defaults to `None`.
    
    deeplink_url : `None, `str`
        Deeplink of the application.
        Defaults to `None`.
    
    description : `None`, `str`
        The description of the application. Defaults to empty string.
        Defaults to `None`.
    
    developers : `None`, `tuple` of ``ApplicationEntity``
        The application's games' developers.
        Defaults to `None`.
    
    eula_id : `int`
        The end-user license agreement's id of the application.
        Defaults to `0` if not applicable.
    
    executables : `None`, `tuple` of ``ApplicationExecutable``
        The application's executables.
        Defaults to `None`.
    
    flags : ``ApplicationFlag``
        The application's public flags.
    
    guild_id : `int`
        If the application is a game sold on Discord, this field tells in which guild it is.
        Defaults to `0` if not applicable.
    
    hook : `bool`
        Defaults to `False`.
    
    install_parameters : `None`, ``ApplicationInstallParameters``
        Settings for the application's default in-app authorization link, if enabled.
    
    icon_hash : `int`
        The application's icon's hash as `uint128`.
    
    icon_type : ``IconType``
        The application's icon's type.
    
    id : `int`
        The application's id.
        Defaults to `0`.
    
    max_participants : `int`
        The maximal amount of users, who can join the application's embedded activity.
        Defaults to `0`.
    
    name : `str`
        The name of the application. Defaults to empty string.
    
    overlay : `bool`
        Defaults to `False`.
    
    overlay_compatibility_hook : `bool`
        Defaults to `False`.
    
    owner : ``ClientUserBase``, ``Team``
        The application's owner.
        Defaults to `ZEROUSER`.
    
    primary_sku_id : `int`
        If the application is a game sold on Discord, this field will be the id of the created `Game SKU`.
    
    privacy_policy_url : `None`, `str`
        The url of the application's privacy policy.
        Defaults to `None`.
    
    publishers : `None`, `tuple` of ``ApplicationEntity``
        A list of the application's games' publishers.
        Defaults to `None`.
    
    role_connection_verification_url : `None`, `str`
        The application's role connection verification entry point
    
    rpc_origins : `None`, `tuple` of `str`
        The application's `rpc` origin urls, if `rpc` is enabled.
        Defaults to `None`.
    
    slug : `None`, `str`
        If this application is a game sold on Discord, this field will be the url slug that links to the store page.
        Defaults to `None`.
    
    splash_hash : `int`
        The application's splash image's hash as `uint128`.
    
    splash_type : ``IconType``
        The application's splash image's type.
    
    tags : `None`, `tuple` of `str`
        Up to 5 tags describing the content and functionality of the application.
        Defaults to `None`.
    
    terms_of_service_url : `None`, `str`
        The url of the application's terms of service.
        Defaults to `None`.
    
    third_party_skus : `None`, `tuple` of ``ThirdPartySKU``
         The application's third party stock keeping units.
         Defaults to `None`.
    
    type : ``ApplicationType``
        The application's type.
    
    verify_key : `None`, `str`
        A base64 encoded key for the Game SDK's `GetTicket`.
        Defaults to `None`.
    
    Notes
    -----
    The instances of the class support weakreferencing.
    """
    __slots__ = (
        'aliases', 'bot_public', 'bot_require_code_grant', 'custom_install_url', 'deeplink_url', 'description',
        'developers', 'eula_id', 'executables', 'flags', 'guild_id', 'hook', 'install_parameters', 'max_participants',
        'name', 'overlay', 'overlay_compatibility_hook', 'owner', 'primary_sku_id', 'privacy_policy_url', 'publishers',
        'role_connection_verification_url', 'rpc_origins', 'slug', 'tags', 'terms_of_service_url', 'third_party_skus',
        'type', 'verify_key'
    )
    
    cover = application_cover
    icon = application_icon
    splash = application_splash
    
    @classmethod
    def _create_empty(cls, application_id):
        """
        Creates an empty application, with it's default attributes set.
        
        Parameters
        ----------
        application_id : `int`
            The application's default identifier.
        
        Returns
        -------
        self : ``Application``
            The created application.
        """
        self = object.__new__(cls)
        
        self.aliases = None
        self.bot_public = BOT_PUBLIC_DEFAULT
        self.bot_require_code_grant = BOT_REQUIRE_CODE_GRANT_DEFAULT
        self.cover_hash = 0
        self.cover_type = ICON_TYPE_NONE
        self.custom_install_url = None
        self.deeplink_url = None
        self.description = None
        self.developers = None
        self.eula_id = 0
        self.executables = None
        self.flags = ApplicationFlag()
        self.guild_id = 0
        self.hook = HOOK_DEFAULT
        self.install_parameters = None
        self.icon_hash = 0
        self.icon_type = ICON_TYPE_NONE
        self.id = application_id
        self.max_participants = MAX_PARTICIPANTS_DEFAULT
        self.name = ''
        self.overlay = OVERLAY_DEFAULT
        self.overlay_compatibility_hook = OVERLAY_COMPATIBILITY_HOOK_DEFAULT
        self.owner = ZEROUSER
        self.primary_sku_id = 0
        self.privacy_policy_url = None
        self.publishers = None
        self.role_connection_verification_url = None
        self.rpc_origins = None
        self.slug = None
        self.splash_hash = 0
        self.splash_type = ICON_TYPE_NONE
        self.tags = None
        self.terms_of_service_url = None
        self.third_party_skus = None
        self.type = ApplicationType.none
        self.verify_key = None
        
        return self
    
    
    @classmethod
    def _base_method_from_data_constructor(cls, self, data):
        """
        Helper method to create the application when calling a base-method from-data constructor.
        
        Parameters
        ----------
        self : `None`, `instance<cls>>`
            The application instance the method was called if any.
        
        data : `dict` of (`str`, `object`) items
            Application data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        application_id = parse_id(data)
        
        if (self is None):
            try:
                self = APPLICATIONS[application_id]
            except KeyError:
                self = cls._create_empty(application_id)
                APPLICATIONS[application_id] = self
        
        elif (self.id == 0):
            try:
                self = APPLICATIONS[application_id]
            except KeyError:
                self.id = application_id
                APPLICATIONS[application_id] = self
        
        elif (self.id != application_id):
            try:
                self = APPLICATIONS[application_id]
            except KeyError:
                self = cls._create_empty(application_id)
                APPLICATIONS[application_id] = self
        
        return self
    
    
    @classmethod
    def _from_data_constructor(cls, data):
        """
        Helper method to create the application when calling a from-data constructor.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Application data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        application_id = parse_id(data)
        
        try:
            self = APPLICATIONS[application_id]
        except KeyError:
            self = cls._create_empty(application_id)
            APPLICATIONS[application_id] = self
        
        return self
    

    @classmethod
    def from_data(cls, data):
        """
        Creates a new application with the given data.
        
        Please use a specialised method instead:
        
        - ``.from_data_ready``
        - ``.from_data_own``
        - ``.from_data_invite``
        - ``.from_data_detectable``
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Application data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        warnings.warn(
            (
                f'`{cls.__name__}` has specialized data-based constructors. They are: `.from_data_ready`, '
                f' `.from_data_own`, `.from_data_invite`, `.from_data_detectable`. Please use own of those instead.'
            ),
            RuntimeWarning,
            stacklevel = 2,
        )
        
        self = cls._from_data_constructor(data)
        self._update_attributes_common(data)
        return self
    
    
    @BaseMethodDescriptor
    def from_data_ready(cls, self, data):
        """
        Creates a new application if the given data refers to an other one. Updates the application and returns it.
        
        This method is called in a `ready` event parser to update the client's application.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Application data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = cls._base_method_from_data_constructor(self, data)
        self._update_attributes_ready(data)
        return self
    
    
    @BaseMethodDescriptor
    def from_data_own(cls, self, data):
        """
        Creates a new application if the given data refers to an other one. Updates the application and returns it.
        
        This method is called when a client updates it's own application.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Application data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = cls._base_method_from_data_constructor(self, data)
        self._update_attributes_own(data)
        return self
    
    
    @classmethod
    def from_data_invite(cls, data):
        """
        Creates an application from the given data. If the application already exists picks it up.
        
        This method is called when an invite contains application data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Application data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = cls._from_data_constructor(data)
        self._update_attributes_invite(data)
        return self
    
    
    @classmethod
    def from_data_detectable(cls, data):
        """
        Creates an application from the given data. If the application already exists picks it up.
        
        This method is called when the detectable applications are requested.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Application data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = cls._from_data_constructor(data)
        self._update_attributes_detectable(data)
        return self
    
    
    def _update_attributes_ready(self, data):
        """
        Updates the application's attributes from a ready event application data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items, Optional
            Application data.
        """
        self.flags = parse_flags(data)
    
    
    def _update_attributes_own(self, data):
        """
        Updates the application's attributes when it's own data was requested.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items, Optional
            Application data.
        """
        self._update_attributes_common(data)
        self.custom_install_url = parse_custom_install_url(data)
        self.guild_id = parse_guild_id(data)
        self.install_parameters = parse_install_parameters(data)
        self.owner = parse_owner(data)
        self.primary_sku_id = parse_primary_sku_id(data)
        self.role_connection_verification_url = parse_role_connection_verification_url(data)
        self.slug = parse_slug(data)
    
    
    def _update_attributes_invite(self, data):
        """
        Updates the application's attributes when it's data is part of an invite.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items, Optional
            Application data.
        """
        self._update_attributes_common(data)
        self.max_participants = parse_max_participants(data)
    
    
    def _update_attributes_detectable(self, data):
        """
        Updates the application's attributes when requesting all detectable applications.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items, Optional
            Application data.
        """
        self._update_attributes_common(data)
        self.aliases = parse_aliases(data)
        self.deeplink_url = parse_deeplink_url(data)
        self.developers = parse_developers(data)
        self.eula_id = parse_eula_id(data)
        self.executables = parse_executables(data)
        self.guild_id = parse_guild_id(data)
        self.overlay = parse_overlay(data)
        self.overlay_compatibility_hook = parse_overlay_compatibility_hook(data)
        self.primary_sku_id = parse_primary_sku_id(data)
        self.publishers = parse_publishers(data)
        self.slug = parse_slug(data)
        self.third_party_skus = parse_third_party_skus(data)
    
    
    def _update_attributes_common(self, data):
        """
        Updates the commonly distributed fields of the application.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items, Optional
            Application data.
        """
        self.bot_public = parse_bot_public(data)
        self.bot_require_code_grant = parse_bot_require_code_grant(data)
        self._set_cover(data)
        self.description = parse_description(data)
        self.flags = parse_flags(data)
        self.hook = parse_hook(data)
        self._set_icon(data)
        self.name = parse_name(data)
        self.privacy_policy_url = parse_privacy_policy_url(data)
        self.rpc_origins = parse_rpc_origins(data)
        self._set_splash(data)
        self.tags = parse_tags(data)
        self.terms_of_service_url = parse_terms_of_service_url(data)
        self.type = parse_type(data)
        self.verify_key = parse_verify_key(data)
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Convert the application back to a json serializable dictionary.
        
        Please use a specialised method instead:
        
        - ``.to_data_ready``
        - ``.to_data_own``
        - ``.to_data_invite``
        - ``.to_data_detectable``
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `str`) items
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}` has specialized to-data converters. They are: `.to_data_ready`, '
                f' `.to_data_own`, `.to_data_invite`, `.to_data_detectable`. Please use own of those instead.'
            ),
            RuntimeWarning,
            stacklevel = 2,
        )
        
        return self._to_data_common(defaults, include_internals)
    
    
    def to_data_ready(self, *, defaults = False, include_internals = False):
        """
        Convert the application to a json serializable dictionary matching a ready event application one.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `str`) items
        """
        data = {}
        put_flags_into(self.flags, data, defaults)
        if include_internals:
            put_id_into(self.id, data, defaults)
        return data
    
    
    def to_data_own(self, *, defaults = False, include_internals = False):
        """
        Convert the application to a json serializable dictionary matching the payload received from requesting the
        client's own payload.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `str`) items
        """
        data = self._to_data_common(defaults, include_internals)
        put_custom_install_url_into(self.custom_install_url, data, defaults)
        put_guild_id_into(self.guild_id, data, defaults)
        put_install_parameters_into(self.install_parameters, data, defaults)
        if include_internals:
            put_owner_into(self.owner, data, defaults)
        put_primary_sku_id_into(self.primary_sku_id, data, defaults)
        put_role_connection_verification_url_into(self.role_connection_verification_url, data, defaults)
        put_slug_into(self.slug, data, defaults)
        return data
    
    
    def to_data_invite(self, *, defaults = False, include_internals = False):
        """
        Convert the application to a json serializable dictionary matching an invite's application data.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `str`) items
        """
        data = self._to_data_common(defaults, include_internals)
        put_max_participants_into(self.max_participants, data, defaults)
        return data
    
    
    def to_data_detectable(self, *, defaults = False, include_internals = False):
        """
        Convert the application to a json serializable dictionary matching a detectable application's data.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `str`) items
        """
        data = self._to_data_common(defaults, include_internals)
        put_aliases_into(self.aliases, data, defaults)
        put_deeplink_url_into(self.deeplink_url, data, defaults)
        put_developers_into(self.developers, data, defaults, include_internals = True)
        put_eula_id_into(self.eula_id, data, defaults)
        put_executables_into(self.executables, data, defaults)
        put_guild_id_into(self.guild_id, data, defaults)
        put_overlay_into(self.overlay, data, defaults)
        put_overlay_compatibility_hook_into(self.overlay_compatibility_hook, data, defaults)
        put_primary_sku_id_into(self.primary_sku_id, data, defaults)
        put_publishers_into(self.publishers, data, defaults, include_internals = True)
        put_slug_into(self.slug, data, defaults)
        put_third_party_skus_into(self.third_party_skus, data, defaults)
        return data
    
    
    def _to_data_common(self, defaults, include_internals):
        """
        Convert the application back to a json serializable dictionary.
        
        Parameters
        ----------
        defaults : `bool`
            Whether default values should be included as well.
        include_internals : `bool`
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `str`) items
        """
        data = {}
        put_bot_public_into(self.bot_public, data, defaults)
        put_bot_require_code_grant_into(self.bot_require_code_grant, data, defaults)
        type(self).cover.put_into(self.cover, data, defaults, as_data = not include_internals)
        put_description_into(self.description, data, defaults)
        put_flags_into(self.flags, data, defaults)
        put_hook_into(self.hook, data, defaults)
        if include_internals:
            put_id_into(self.id, data, defaults)
        type(self).icon.put_into(self.icon, data, defaults, as_data = not include_internals)
        put_name_into(self.name, data, defaults)
        put_privacy_policy_url_into(self.privacy_policy_url, data, defaults)
        put_rpc_origins_into(self.rpc_origins, data, defaults)
        type(self).splash.put_into(self.splash, data, defaults, as_data = not include_internals)
        put_tags_into(self.tags, data, defaults)
        put_terms_of_service_url_into(self.terms_of_service_url, data, defaults)
        put_type_into(self.type, data, defaults)
        put_verify_key_into(self.verify_key, data, defaults)
        return data
    
    
    def _create_update(self, data, ready_data):
        """
        Creates a new application if the given data refers to an other one. Updates the application and returns it.
        
        Deprecated and will be removed in 2023 Marc.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items, Optional
            Application data received from Discord.
        ready_data : `bool`
            Whether the application data was received from a ready event.
        
        Returns
        -------
        self : ``Application``
            The created or updated application.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.embedded_activity_configuration` is deprecated and will be removed in '
                f'2023 Marc.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        if ready_data:
            self = self.from_data_ready(data)
        else:
            self = self.from_data_own(data)
        return self
    
    
    def __hash__(self):
        """Returns the application's hash value."""
        application_id = self.id
        if application_id:
            return application_id
        
        return self._get_hash_partial()
    
        
    def _get_hash_partial(self):
        """
        Hashes the attributes of the application.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # aliases
        aliases = self.aliases
        if (aliases is not None):
            hash_value ^= len(aliases) << 0
            
            for alias in aliases:
                hash_value ^= hash(alias)
        
        # bot_public
        hash_value ^= self.bot_public << 4
        
        # bot_require_code_grant
        hash_value ^= self.bot_require_code_grant << 5
        
        # cover
        hash_value ^= hash(self.cover)
        
        # custom_install_url
        custom_install_url = self.custom_install_url
        if (custom_install_url is not None):
            hash_value ^= hash(custom_install_url)
        
        # deeplink_url
        deeplink_url = self.deeplink_url
        if (deeplink_url is not None):
            hash_value ^= hash(deeplink_url)
        
        description = self.description
        if (description is not None) and (description != self.name):
            hash_value ^= hash(description)
        
        # developers
        developers = self.developers
        if (developers is not None):
            hash_value ^= len(developers) << 6
            
            for developer in developers:
                hash_value ^= hash(developer)
        
        # eula_id
        hash_value ^= self.eula_id
        
        # executables
        executables = self.executables
        if (executables is not None):
            hash_value ^= len(executables) << 10
            
            for executable in executables:
                hash_value ^= hash(executable)
        
        # flags
        hash_value ^= self.flags
        
        # guild_id
        hash_value ^= hash(self.guild_id)
        
        # hook
        hash_value ^= self.hook << 14
        
        # install_parameters
        install_parameters = self.install_parameters
        if (install_parameters is not None):
            hash_value ^= hash(install_parameters)
        
        # icon
        hash_value ^= hash(self.icon)
        
        # max_participants
        hash_value ^= self.max_participants << 15
        
        # name
        hash_value ^= hash(self.name)
        
        # overlay
        hash_value ^= self.overlay << 19
        
        # overlay_compatibility_hook
        hash_value ^= self.overlay_compatibility_hook << 20
        
        # owner
        owner = self.owner
        if (owner is not ZEROUSER):
            hash_value ^= hash(owner)
        
        # primary_sku_id
        hash_value ^= self.primary_sku_id
        
        # privacy_policy_url
        privacy_policy_url = self.privacy_policy_url
        if (privacy_policy_url is not None):
            hash_value ^= hash(privacy_policy_url)
        
        # publishers
        publishers = self.publishers
        if (publishers is not None):
            hash_value ^= len(publishers) << 21
            
            for publisher in publishers:
                hash_value ^= hash(publisher)
        
        # role_connection_verification_url
        role_connection_verification_url = self.role_connection_verification_url
        if (role_connection_verification_url is not None):
            hash_value ^= hash(role_connection_verification_url)
        
        # rpc_origins
        rpc_origins = self.rpc_origins
        if (rpc_origins is not None):
            hash_value ^= len(rpc_origins) << 25
            
            for rpc_origin in rpc_origins:
                hash_value ^= hash(rpc_origin)
        
        # slug
        slug = self.slug
        if (slug is not None):
            hash_value ^= hash(slug)
        
        # splash
        hash_value ^= hash(self.splash)
        
        # tags
        tags = self.tags
        if (tags is not None):
            hash_value ^= len(tags) << 29
            
            for tag in tags:
                hash_value ^= hash(tag)
        
        # terms_of_service_url
        terms_of_service_url = self.terms_of_service_url
        if (terms_of_service_url is not None):
            hash_value ^= hash(terms_of_service_url)
        
        # third_party_skus
        third_party_skus = self.third_party_skus
        if (third_party_skus is not None):
            hash_value ^= len(third_party_skus) << 1
            
            for third_party_sku in third_party_skus:
                hash_value ^= hash(third_party_sku)
        
        # type
        hash_value ^= hash(self.type) << 5
        
        # verify_key
        verify_key = self.verify_key
        if (verify_key is not None):
            hash_value ^= hash(verify_key)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two applications are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two applications are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether self is equal to other. Other must be same type as self.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        is_equal : `bool`
        """
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            if self.id != other.id:
                return False
        
        # aliases
        if self.aliases != other.aliases:
            return False
        
        # bot_public
        if self.bot_public != other.bot_public:
            return False
        
        # bot_require_code_grant
        if self.bot_require_code_grant != other.bot_require_code_grant:
            return False
        
        # cover_hash
        if self.cover_hash != other.cover_hash:
            return False
        
        # cover_type
        if self.cover_type != other.cover_type:
            return False
        
        # custom_install_url
        if self.custom_install_url != other.custom_install_url:
            return False
        
        # deeplink_url
        if self.deeplink_url != other.deeplink_url:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # developers
        if self.developers != other.developers:
            return False
        
        # eula_id
        if self.eula_id != other.eula_id:
            return False
        
        # executables
        if self.executables != other.executables:
            return False
        
        # flags
        if self.flags != other.flags:
            return False
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        # hook
        if self.hook != other.hook:
            return False
        
        # install_parameters
        if self.install_parameters != other.install_parameters:
            return False
        
        # icon_hash
        if self.icon_hash != other.icon_hash:
            return False
        
        # icon_type
        if self.icon_type != other.icon_type:
            return False
        
        # max_participants
        if self.max_participants != other.max_participants:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # overlay
        if self.overlay != other.overlay:
            return False
        
        # overlay_compatibility_hook
        if self.overlay_compatibility_hook != other.overlay_compatibility_hook:
            return False
        
        # owner
        if self.owner != other.owner:
            return False
        
        # primary_sku_id
        if self.primary_sku_id != other.primary_sku_id:
            return False
        
        # privacy_policy_url
        if self.privacy_policy_url != other.privacy_policy_url:
            return False
        
        # publishers
        if self.publishers != other.publishers:
            return False
        
        # role_connection_verification_url
        if self.role_connection_verification_url != other.role_connection_verification_url:
            return False
        
        # rpc_origins
        if self.rpc_origins != other.rpc_origins:
            return False
        
        # slug
        if self.slug != other.slug:
            return False
        
        # splash_hash
        if self.splash_hash != other.splash_hash:
            return False
        
        # splash_type
        if self.splash_type != other.splash_type:
            return False
        
        # tags
        if self.tags != other.tags:
            return False
        
        # terms_of_service_url
        if self.terms_of_service_url != other.terms_of_service_url:
            return False
        
        # third_party_skus
        if self.third_party_skus != other.third_party_skus:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # verify_key
        if self.verify_key != other.verify_key:
            return False
        
        return True
    
    
    def __repr__(self):
        """Returns the application's representation"""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        application_id = self.id
        if application_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(application_id))
            repr_parts.append(',')
        else:
            repr_parts.append(' (partial)')
    
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __new__(cls, **keyword_parameters):
        """
        Creates a partial application with the given parameters.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            Additional parameters defining the attributes of the application.
        
        Other parameters
        ----------------
        aliases : `None`, `iterable` of `str`, Optional (Keyword only)
            Aliases of the application's name.
        
        application_type : `int`, ``ApplicationType``, Optional (Keyword only)
            The application's type.
        
        bot_public : `bool`, Optional (Keyword only)
            Whether not only the application's owner can join the application's bot to guilds.
            
        bot_require_code_grant : `bool`, Optional (Keyword only)
            Whether the application's bot will only join a guild, when completing the full `oauth2` code grant flow.
        
        cover : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The application's cover.
        
        custom_install_url : `None`, `str`, Optional (Keyword only)
            The application's default custom authorization link if enabled.
        
        deeplink_url : `None, `str`, Optional (Keyword only)
            Deeplink of the application.
        
        description : `None`, `str`, Optional (Keyword only)
            The description of the application.
        
        developers : `None`, `iterable` of ``ApplicationEntity``, Optional (Keyword only)
            The application's games' developers.
        
        eula_id : `int`, Optional (Keyword only)
            The end-user license agreement's id of the application.
        
        executables : `None`, `iterable` of ``ApplicationExecutable``, Optional (Keyword only)
            The application's executables.
        
        flags : `int`, ``ApplicationFlag``, Optional (Keyword only)
            The application's public flags.
        
        guild_id : `int`, Optional (Keyword only)
            If the application is a game sold on Discord, this field tells in which guild it is.
        
        hook : `bool`, Optional (Keyword only)
            ???
        
        install_parameters : `None`, ``ApplicationInstallParameters``, Optional (Keyword only)
            Settings for the application's default in-app authorization link, if enabled.
        
        icon : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The application's icon.
        
        max_participants : `int`, Optional (Keyword only)
            The maximal amount of users, who can join the application's embedded activity.
        
        name : `str`, Optional (Keyword only)
            The name of the application. Defaults to empty string.
        
        overlay : `bool`, Optional (Keyword only)
            ???
        
        overlay_compatibility_hook : `bool`, Optional (Keyword only)
            ???
        
        owner : ``ClientUserBase``, ``Team``, Optional (Keyword only)
            The application's owner.
        
        primary_sku_id : `int`, Optional (Keyword only)
            If the application is a game sold on Discord, this field will be the id of the created `Game SKU`.
        
        privacy_policy_url : `None`, `str`, Optional (Keyword only)
            The url of the application's privacy policy.
        
        publishers : `None`, `iterable` of ``ApplicationEntity``, Optional (Keyword only)
            A list of the application's games' publishers.
        
        role_connection_verification_url : `None`, `str`, Optional (Keyword only)
            The application's role connection verification entry point
        
        rpc_origins : `None`, `iterable` of `str`, Optional (Keyword only)
            The application's `rpc` origin urls, if `rpc` is enabled.
        
        slug : `None`, `str`, Optional (Keyword only)
            If this application is a game sold on Discord, this field will be the url slug that links to the store page.
        
        splash : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The application's splash.
        
        tags : `None`, `iterable` of `str`, Optional (Keyword only)
            Up to 5 tags describing the content and functionality of the application.
        
        terms_of_service_url : `None`, `str`, Optional (Keyword only)
            The url of the application's terms of service.
        
        third_party_skus : `None`, `iterable` of ``ThirdPartySKU``, Optional (Keyword only)
             The application's third party stock keeping units.
        
        verify_key : `None`, `str`, Optional (Keyword only)
            A base64 encoded key for the Game SDK's `GetTicket`.
        
        Returns
        -------
        application : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra parameter(s).
        ValueError
            - If an parameter's value is incorrect.
        """
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, NEW_FIELDS)
        else:
            processed = None
        
        self = cls._create_empty(0)
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    @classmethod
    def precreate(cls, application_id, **keyword_parameters):
        """
        Precreates an application with the given parameters.
        
        Parameters
        ----------
        application_id : `int`, `str`
            The application's id.
        
        **keyword_parameters : Keyword parameters
            Additional parameters defining the attributes of the application.
        
        Other parameters
        ----------------
        aliases : `None`, `iterable` of `str`, Optional (Keyword only)
            Aliases of the application's name.
        
        application_type : `int`, ``ApplicationType``, Optional (Keyword only)
            The application's type.
        
        bot_public : `bool`, Optional (Keyword only)
            Whether not only the application's owner can join the application's bot to guilds.
            
        bot_require_code_grant : `bool`, Optional (Keyword only)
            Whether the application's bot will only join a guild, when completing the full `oauth2` code grant flow.
        
        cover : `None`, ``Icon``, `str`, Optional (Keyword only)
            The application's cover.
        
        custom_install_url : `None`, `str`, Optional (Keyword only)
            The application's default custom authorization link if enabled.
        
        deeplink_url : `None, `str`, Optional (Keyword only)
            Deeplink of the application.
        
        description : `None`, `str`, Optional (Keyword only)
            The description of the application.
        
        developers : `None`, `iterable` of ``ApplicationEntity``, Optional (Keyword only)
            The application's games' developers.
        
        eula_id : `int`, Optional (Keyword only)
            The end-user license agreement's id of the application.
        
        executables : `None`, `iterable` of ``ApplicationExecutable``, Optional (Keyword only)
            The application's executables.
        
        flags : `int`, ``ApplicationFlag``, Optional (Keyword only)
            The application's public flags.
        
        guild_id : `int`, Optional (Keyword only)
            If the application is a game sold on Discord, this field tells in which guild it is.
        
        hook : `bool`, Optional (Keyword only)
            ???
        
        install_parameters : `None`, ``ApplicationInstallParameters``, Optional (Keyword only)
            Settings for the application's default in-app authorization link, if enabled.
        
        icon : `None`, ``Icon``, `str`, Optional (Keyword only)
            The application's icon.
        
        max_participants : `int`, Optional (Keyword only)
            The maximal amount of users, who can join the application's embedded activity.
        
        name : `str`, Optional (Keyword only)
            The name of the application. Defaults to empty string.
        
        overlay : `bool`, Optional (Keyword only)
            ???
        
        overlay_compatibility_hook : `bool`, Optional (Keyword only)
            ???
        
        owner : ``ClientUserBase``, ``Team``, Optional (Keyword only)
            The application's owner.
        
        primary_sku_id : `int`, Optional (Keyword only)
            If the application is a game sold on Discord, this field will be the id of the created `Game SKU`.
        
        privacy_policy_url : `None`, `str`, Optional (Keyword only)
            The url of the application's privacy policy.
        
        publishers : `None`, `iterable` of ``ApplicationEntity``, Optional (Keyword only)
            A list of the application's games' publishers.
        
        role_connection_verification_url : `None`, `str`, Optional (Keyword only)
            The application's role connection verification entry point
        
        rpc_origins : `None`, `iterable` of `str`, Optional (Keyword only)
            The application's `rpc` origin urls, if `rpc` is enabled.
        
        slug : `None`, `str`, Optional (Keyword only)
            If this application is a game sold on Discord, this field will be the url slug that links to the store page.
        
        splash : `None`, ``Icon``, `str`, Optional (Keyword only)
            The application's splash.
        
        tags : `None`, `iterable` of `str`, Optional (Keyword only)
            Up to 5 tags describing the content and functionality of the application.
        
        terms_of_service_url : `None`, `str`, Optional (Keyword only)
            The url of the application's terms of service.
        
        third_party_skus : `None`, `iterable` of ``ThirdPartySKU``, Optional (Keyword only)
             The application's third party stock keeping units.
        
        verify_key : `None`, `str`, Optional (Keyword only)
            A base64 encoded key for the Game SDK's `GetTicket`.
        
        
        Returns
        -------
        application : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra parameter(s).
        ValueError
            - If an parameter's value is incorrect.
        """
        application_id = validate_id(application_id)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        try:
            self = APPLICATIONS[application_id]
        except KeyError:
            self = cls._create_empty(application_id)
            APPLICATIONS[application_id] = self
            update = True
        else:
            update = self.partial
        
        if update and (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    def copy_with(self, **keyword_parameters):
        """
        Copies the application with the given attributes replaced.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            Additional parameters defining the attributes of the copy.
        
        Other parameters
        ----------------
        aliases : `None`, `iterable` of `str`, Optional (Keyword only)
            Aliases of the application's name.
        
        application_type : `int`, ``ApplicationType``, Optional (Keyword only)
            The application's type.
        
        bot_public : `bool`, Optional (Keyword only)
            Whether not only the application's owner can join the application's bot to guilds.
            
        bot_require_code_grant : `bool`, Optional (Keyword only)
            Whether the application's bot will only join a guild, when completing the full `oauth2` code grant flow.
        
        cover : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The application's cover.
        
        custom_install_url : `None`, `str`, Optional (Keyword only)
            The application's default custom authorization link if enabled.
        
        deeplink_url : `None, `str`, Optional (Keyword only)
            Deeplink of the application.
        
        description : `None`, `str`, Optional (Keyword only)
            The description of the application.
        
        developers : `None`, `iterable` of ``ApplicationEntity``, Optional (Keyword only)
            The application's games' developers.
        
        eula_id : `int`, Optional (Keyword only)
            The end-user license agreement's id of the application.
        
        executables : `None`, `iterable` of ``ApplicationExecutable``, Optional (Keyword only)
            The application's executables.
        
        flags : `int`, ``ApplicationFlag``, Optional (Keyword only)
            The application's public flags.
        
        guild_id : `int`, Optional (Keyword only)
            If the application is a game sold on Discord, this field tells in which guild it is.
        
        hook : `bool`, Optional (Keyword only)
            ???
        
        install_parameters : `None`, ``ApplicationInstallParameters``, Optional (Keyword only)
            Settings for the application's default in-app authorization link, if enabled.
        
        icon : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The application's icon.
        
        max_participants : `int`, Optional (Keyword only)
            The maximal amount of users, who can join the application's embedded activity.
        
        name : `str`, Optional (Keyword only)
            The name of the application. Defaults to empty string.
        
        overlay : `bool`, Optional (Keyword only)
            ???
        
        overlay_compatibility_hook : `bool`, Optional (Keyword only)
            ???
        
        owner : ``ClientUserBase``, ``Team``, Optional (Keyword only)
            The application's owner.
        
        primary_sku_id : `int`, Optional (Keyword only)
            If the application is a game sold on Discord, this field will be the id of the created `Game SKU`.
        
        privacy_policy_url : `None`, `str`, Optional (Keyword only)
            The url of the application's privacy policy.
        
        publishers : `None`, `iterable` of ``ApplicationEntity``, Optional (Keyword only)
            A list of the application's games' publishers.
        
        role_connection_verification_url : `None`, `str`, Optional (Keyword only)
            The application's role connection verification entry point
        
        rpc_origins : `None`, `iterable` of `str`, Optional (Keyword only)
            The application's `rpc` origin urls, if `rpc` is enabled.
        
        slug : `None`, `str`, Optional (Keyword only)
            If this application is a game sold on Discord, this field will be the url slug that links to the store page.
        
        splash : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The application's splash.
        
        tags : `None`, `iterable` of `str`, Optional (Keyword only)
            Up to 5 tags describing the content and functionality of the application.
        
        terms_of_service_url : `None`, `str`, Optional (Keyword only)
            The url of the application's terms of service.
        
        third_party_skus : `None`, `iterable` of ``ThirdPartySKU``, Optional (Keyword only)
             The application's third party stock keeping units.
        
        verify_key : `None`, `str`, Optional (Keyword only)
            A base64 encoded key for the Game SDK's `GetTicket`.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra parameter(s).
        ValueError
            - If an parameter's value is incorrect.
        """
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, NEW_FIELDS)
        else:
            processed = None
        
        new = self.copy()
        
        if (processed is not None):
            for item in processed:
                setattr(new, *item)
        
        return new
    
    
    def copy(self):
        """
        Copies the application returning a new partial one.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        aliases = self.aliases
        if (aliases is not None):
            aliases = (*aliases,)
        new.aliases = aliases
        new.bot_public = self.bot_public
        new.bot_require_code_grant = self.bot_require_code_grant
        new.cover_hash = self.cover_hash
        new.cover_type = self.cover_type
        new.custom_install_url = self.custom_install_url
        new.deeplink_url = self.deeplink_url
        new.description = self.description
        developers = self.developers
        if (developers is not None):
            developers = (*(developer.copy() for developer in developers),)
        new.developers = developers
        new.eula_id = self.eula_id
        executables = self.executables
        if (executables is not None):
            executables = (*(executable.copy() for executable in executables),)
        new.executables = executables
        new.flags = self.flags
        new.guild_id = self.guild_id
        new.hook = self.hook
        install_parameters = self.install_parameters
        if (install_parameters is not None):
            install_parameters = install_parameters.copy()
        new.install_parameters = install_parameters
        new.icon_hash = self.icon_hash
        new.icon_type = self.icon_type
        new.id = 0
        new.max_participants = self.max_participants
        new.name = self.name
        new.overlay = self.overlay
        new.overlay_compatibility_hook = self.overlay_compatibility_hook
        new.owner = self.owner # Do not copy ~ yet
        new.primary_sku_id = self.primary_sku_id
        new.privacy_policy_url = self.privacy_policy_url
        publishers = self.publishers
        if (publishers is not None):
            publishers = (*(publisher.copy() for publisher in publishers),)
        new.publishers = publishers
        new.role_connection_verification_url = self.role_connection_verification_url
        rpc_origins = self.rpc_origins
        if (rpc_origins is not None):
            rpc_origins = (*rpc_origins,)
        new.rpc_origins = rpc_origins
        new.slug = self.slug
        new.splash_hash = self.splash_hash
        new.splash_type = self.splash_type
        tags = self.tags
        if (tags is not None):
            tags = (*tags,)
        new.tags = tags
        new.terms_of_service_url = self.terms_of_service_url
        third_party_skus = self.third_party_skus
        if (third_party_skus is not None):
            third_party_skus = (*(third_party_sku.copy() for third_party_sku in third_party_skus),)
        new.third_party_skus = third_party_skus
        new.type = self.type
        new.verify_key = self.verify_key
        return new
    
    
    @property
    def partial(self):
        """
        Returns whether the application is partial.
        
        Returns
        -------
        partial : `bool`
        """
        application_id = self.id
        if not application_id:
            return True
        
        try:
            client = APPLICATION_ID_TO_CLIENT[application_id]
        except KeyError:
            return True
        
        if not client.running:
            return True
        
        return False
    
    
    @property
    def embedded_activity_configuration(self):
        """
        Deprecated and will be removed in 2023 Marc.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.embedded_activity_configuration` is deprecated and will be removed in '
                f'2023 Marc.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return None
    
    
    def iter_aliases(self):
        """
        Iterates over the aliases of the application.
        
        This method is an iterable generator.
        
        Yields
        ------
        alias : `str`
        """
        aliases = self.aliases
        if (aliases is not None):
            yield from aliases
    
    
    def iter_developers(self):
        """
        Iterates over the developers of the application.
        
        This method is an iterable generator.
        
        Yields
        ------
        developer : ``ApplicationEntity``
        """
        developers = self.developers
        if (developers is not None):
            yield from developers
    
    
    def iter_executables(self):
        """
        Iterates over the executables of the application.
        
        This method is an iterable generator.
        
        Yields
        ------
        executable : ``ApplicationExecutable``
        """
        executables = self.executables
        if (executables is not None):
            yield from executables
    
    
    def iter_publishers(self):
        """
        Iterates over the publishers of the application.
        
        This method is an iterable generator.
        
        Yields
        ------
        publisher : ``ApplicationEntity``
        """
        publishers = self.publishers
        if (publishers is not None):
            yield from publishers
    
    
    def iter_rpc_origins(self):
        """
        Iterates over the rpc origins of the application.
        
        This method is an iterable generator.
        
        Yields
        ------
        rpc_origin : `str`
        """
        rpc_origins = self.rpc_origins
        if (rpc_origins is not None):
            yield from rpc_origins
    
    
    def iter_tags(self):
        """
        Iterates over the tags of the application.
        
        This method is an iterable generator.
        
        Yields
        ------
        alias : `str`
        """
        tags = self.tags
        if (tags is not None):
            yield from tags
    
    
    def iter_third_party_skus(self):
        """
        Iterates over the third party sku-s of the application.
        
        This method is an iterable generator.
        
        Yields
        ------
        third_party_sku : ``ThirdPartySKU``
        """
        third_party_skus = self.third_party_skus
        if (third_party_skus is not None):
            yield from third_party_skus



__all__ = ()

from scarletio import Future, RichAttributeErrorBaseType

from ..core import KOKORO


class SoundboardSoundsEventWaiter(RichAttributeErrorBaseType):
    """
    Soundboard sounds event waiter.
    
    Attributes
    ----------
    counter : `int`
        Counter how much times the event is waited on.
    future : ``Future``
        The waiter future.
    """
    __slots__ = ('counter', 'future')
    
    
    def __new__(cls):
        """
        Creates a new event waiter.
        """
        self = object.__new__(cls)
        self.counter = 1
        self.future = Future(KOKORO)
        return self
    
    
    def __repr__(self):
        """Returns the event waiter's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' counter = ')
        repr_parts.append(repr(self.counter))
        
        repr_parts.append(', future = ')
        repr_parts.append(repr(self.future))
        
        repr_parts.append('>')
        return ''.join(repr_parts)



__all__ = ('TeamMemberPermission', 'TeamMembershipState')

from ...bases import Preinstance as P, PreinstancedBase


class TeamMembershipState(PreinstancedBase):
    """
    Represents a ``TeamMember``'s state at a ``Team``.
    
    Attributes
    ----------
    name : `str`
        The name of state.
    value : `int`
        The Discord side identifier value of the team membership state.
        
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``TeamMembershipState``) items
        Stores the created team membership state instances. This container is accessed when translating a Discord
        team membership state's value to it's representation.
    VALUE_TYPE : `type` = `int`
        The team membership states' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the team membership states.
    
    Every predefined team membership state can be accessed as class attribute as well:
    +-----------------------+-----------+-------+
    | Class attribute name  | name      | value |
    +=======================+===========+=======+
    | none                  | none      | 0     |
    +-----------------------+-----------+-------+
    | invited               | invited   | 1     |
    +-----------------------+-----------+-------+
    | accepted              | accepted  | 2     |
    +-----------------------+-----------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    invited = P(1, 'invited')
    accepted = P(2, 'accepted')


class TeamMemberPermission(PreinstancedBase):
    """
    Represents a permission of a ``TeamMember``.
    
    Attributes
    ----------
    name : `str`
        The name of permission.
    value : `str`
        The Discord side identifier value of the team member permission.
        
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``TeamMemberPermission``) items
        Stores the created team member permission instances. This container is accessed when translating a Discord
        team member permission's value to it's representation.
    VALUE_TYPE : `type` = `str`
        The team member permissions' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the team member permissions.
    
    Every predefined team member permission can be accessed as class attribute as well:
    +-----------------------+-----------+-------+
    | Class attribute name  | name      | value |
    +=======================+===========+=======+
    | none                  | none      | `''`  |
    +-----------------------+-----------+-------+
    | admin                 | admin     | `'*'` |
    +-----------------------+-----------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = str
    
    __slots__ = ()
    
    # predefined
    none = P('', 'none')
    admin = P('*', 'admin')



__all__ = ('IntegrationType', )

from scarletio import export

from ...bases import Preinstance as P, PreinstancedBase

from ..integration_metadata import IntegrationMetadataBase, IntegrationMetadataSubscription, IntegrationMetadataDiscord


@export
class IntegrationType(PreinstancedBase):
    """
    Represents an ``Integration``'s type.
    
    Attributes
    ----------
    name : `str`
        The name of the integration type.
    value : `int`
        The Discord side identifier value of the integration type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``IntegrationType``) items
        Stores the predefined ``IntegrationType``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The integration type' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the integration types.
    
    Every predefined integration type can be accessed as class attribute as well:
    
    +-----------------------+-----------------------+---------------------------+---------------------------------------+
    | Class attribute name  | Name                  | Value                     | Metadata type                         |
    +=======================+=======================+===========================+=======================================+
    | none                  | none                  | `''`                      | ``IntegrationMetadataBase``           |
    +-----------------------+-----------------------+---------------------------+---------------------------------------+
    | discord               | Discord               | `'discord'`               | ``IntegrationMetadataDiscord``        |
    +-----------------------+-----------------------+---------------------------+---------------------------------------+
    | twitch                | Twitch                | `'twitch'`                | ``IntegrationMetadataSubscription``   |
    +-----------------------+-----------------------+---------------------------+---------------------------------------+
    | youtube               | Youtube               | `'youtube'`               | ``IntegrationMetadataSubscription``   |
    +-----------------------+-----------------------+---------------------------+---------------------------------------+
    | guild_subscription    | guild subscription    | `'guild_subscription'`    | ``IntegrationMetadataSubscription``   |
    +-----------------------+-----------------------+---------------------------+---------------------------------------+
    
    """
    INSTANCES = {}
    VALUE_TYPE = str
    
    __slots__ = ('metadata_type',)
    

    @classmethod
    def _from_value(cls, value):
        """
        Creates a new integration type with the given value.
        
        Parameters
        ----------
        value : `str`
            The integration's type.
        
        Returns
        -------
        self : ``IntegrationType``
            The created instance.
        """
        self = object.__new__(cls)
        self.name = value
        self.value = value
        self.metadata_type = IntegrationMetadataSubscription
        return self
    

    
    def __init__(self, value, name, metadata_type):
        """
        Creates a new scheduled event entity type instance from the given parameters.
        
        Parameters
        ----------
        value : `str`
            The unique identifier of the scheduled event entity type.
        name : `str`
            The name of the scheduled event entity type.
        metadata_type : `None`, ``IntegrationMetadataBase`` subclass
            The integration's metadata's type.
        """
        self.name = name
        self.value = value
        self.metadata_type = metadata_type
        self.INSTANCES[value] = self
    
    
    # predefined
    none = P('', 'none', IntegrationMetadataBase)
    discord = P('discord', 'Discord', IntegrationMetadataDiscord)
    twitch = P('twitch', 'twitch', IntegrationMetadataSubscription)
    youtube = P('youtube', 'Youtube', IntegrationMetadataSubscription)
    guild_subscription = P('guild_subscription', 'guild subscription', IntegrationMetadataSubscription)



__all__ = ()

from scarletio import Compound

from ...application import Application, ApplicationRoleConnectionMetadata
from ...http import DiscordHTTPClient

from .application_command import _assert__application_id


class ClientCompoundApplicationRoleConnectionEndpoints(Compound):
    
    application: Application
    http : DiscordHTTPClient
    
    async def application_role_connection_metadata_get_all(self):
        """
        Requests all the role connection metadatas of the client's application.
        
        Returns
        -------
        application_role_connection_metadatas : `list` of `ApplicationRoleConnectionMetadata`
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        datas = await self.http.application_role_connection_metadata_get_all(application_id)
        return [ApplicationRoleConnectionMetadata.from_data(data) for data in datas]
    
    
    async def application_role_connection_metadata_edit_all(self, application_role_connection_metadatas):
        """
        Edits all the role connection metadatas of the client's application. Basically overwrites the old ones.
        
        Parameters
        ----------
        application_role_connection_metadatas : `iterable` of ``ApplicationRoleConnectionMetadata``
            The new application connection metadatas to replace the old ones.
        
        Returns
        -------
        application_role_connection_metadatas : `list` of `ApplicationRoleConnectionMetadata`
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        datas = [
            application_role_connection_metadata.to_data()
            for application_role_connection_metadata in application_role_connection_metadatas
        ]
        
        datas = await self.http.application_role_connection_metadata_edit_all(application_id, datas)
        return [ApplicationRoleConnectionMetadata.from_data(data) for data in datas]



__all__ = ('EmbeddedActivityState',)

from scarletio import RichAttributeErrorBaseType

from ...activity import Activity
from ...channel import ChannelType, create_partial_channel_from_id
from ...core import EMBEDDED_ACTIVITY_STATES, GUILDS
from ...user import create_partial_user_from_id

from .constants import ACTIVITY_KEY
from .fields import (
    parse_activity, parse_user_ids, put_activity_into, put_channel_id_into, put_guild_id_into, put_user_ids_into,
    validate_activity, validate_channel_id, validate_guild_id, validate_user_ids
)
from .helpers import _add_embedded_activity_state_to_guild_cache, _remove_embedded_activity_state_from_guild_cache
from .key import EmbeddedActivityStateKey


class EmbeddedActivityState(RichAttributeErrorBaseType):
    """
    Represents an embedded activity's state.
    
    Attributes
    ----------
    activity : ``Activity``
        The embedded activity.
    channel_id : `int`
        The respective channel's identifier.
    guild_id : `int`
        The respective guild's identifier.
    user_ids : `set` of `int`
        The joined users' identifiers.
    
    Notes
    -----
    Embedded activity state instances are weakreferable.
    """
    __slots__ = ('__weakref__', 'activity', 'channel_id', 'guild_id', 'user_ids')
    
    
    def __new__(cls, *, activity = ..., channel_id = ..., guild_id = ..., user_ids = ...):
        """
        Creates a new partial embedded activity state with the given fields.
        
        Parameters
        ----------
        activity : ``Activity``, Optional (Keyword only)
            The embedded activity.
        channel_id : `int`, ``Channel``, Optional (Keyword only)
            The respective channel or its identifier.
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The respective guild or its identifier.
        user_ids : `None`, `iterable` of (`int`, ``ClientUserBase``), Optional (Keyword only)
            The joined users' identifiers.
        
        Raises
        ------
        TypeError
            - If a parameter's value is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # activity
        if activity is ...:
            activity = Activity()
        else:
            activity = validate_activity(activity)
        
        # channel_id
        if channel_id is ...:
            channel_id = 0
        else:
            channel_id = validate_channel_id(channel_id)
        
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_guild_id(guild_id)
        
        # user_ids
        if user_ids is ...:
            user_ids = set()
        else:
            user_ids = validate_user_ids(user_ids)
        
        self = object.__new__(cls)
        self.activity = activity
        self.channel_id = channel_id
        self.guild_id = guild_id
        self.user_ids = user_ids
        return self
        
        
    @classmethod
    def from_data(cls, data, guild_id = 0, *, strong_cache = True):
        """
        Creates a new embedded activity state from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Embedded activity data.
        guild_id : `int` = `0`, Optional
            The guild's identifier where the activity is.
        
        Returns
        -------
        self : `instance<cls>`
        """
        key = EmbeddedActivityStateKey.from_data(data, guild_id)
        
        try:
            self = EMBEDDED_ACTIVITY_STATES[key]
        except KeyError:
            self = cls._from_data_construct(data, key, strong_cache)
        
        return self
    
    
    @classmethod
    def from_data_is_created(cls, data, guild_id = 0):
        """
        Creates a new embedded activity state from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Embedded activity data.
        guild_id : `int` = `0`, Optional
            The guild's identifier where the activity is.
        
        Returns
        -------
        self : `instance<cls>`
        is_created : `bool`
        """
        key = EmbeddedActivityStateKey.from_data(data, guild_id)
        
        try:
            self = EMBEDDED_ACTIVITY_STATES[key]
        except KeyError:
            self = cls._from_data_construct(data, key, True)
            is_created = True
        else:
            is_created = False
        
        return self, is_created
    
    
    @classmethod
    def _from_data_construct(cls, data, key, strong_cache):
        """
        Constructs the embedded activity state from the given fields.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Embedded activity data.
        key : ``EmbeddedActivityStateKey``
            Embedded activity key.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.activity = parse_activity(data)
        self.channel_id = key.channel_id
        self.guild_id = key.guild_id
        self.user_ids = user_ids = parse_user_ids(data)
        
        EMBEDDED_ACTIVITY_STATES[key] = self
        
        if strong_cache and user_ids:
            _add_embedded_activity_state_to_guild_cache(self)
        
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the embedded activity state into it's json serializable representation.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_activity_into(self.activity, data, defaults)
        put_channel_id_into(self.channel_id, data, defaults)
        put_guild_id_into(self.guild_id, data, defaults)
        put_user_ids_into(self.user_ids, data, defaults)
        return data
    
    
    def _update_user_ids(self, data):
        """
        Updates the user id-s of the embedded activity state.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Embedded activity state data.
        """
        self.user_ids = user_ids = parse_user_ids(data)
        
        if not user_ids:
            _remove_embedded_activity_state_from_guild_cache(self)
    
    
    def _difference_update_user_ids(self, data):
        """
        Difference updates the user id-s of the embedded activity state.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Embedded activity state data.
        
        Returns
        -------
        joined_user_ids : `set` of `int`
            The joined users' identifiers.
        left_user_ids : `set` of `int`
            The left users' identifiers.
        """
        new_user_ids = parse_user_ids(data)
        old_user_ids = self.user_ids
        self.user_ids = new_user_ids
        
        if not new_user_ids:
            _remove_embedded_activity_state_from_guild_cache(self)
        
        return new_user_ids - old_user_ids, old_user_ids - new_user_ids
    
    
    def _update_activity(self, data):
        """
        Updates the activity of the embedded activity state.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Embedded activity state data.
        """
        self.activity._update_attributes(data[ACTIVITY_KEY])
    
    
    def _difference_update_activity(self, data):
        """
        Updates the activity of the embedded activity state and returns it's changed attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Embedded activity state data.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `object`) items
            The old attributes of the activity, that were modified.
            Check ``Activity._difference_update_attributes`` for the exact fields.
        """
        return self.activity._difference_update_attributes(data[ACTIVITY_KEY])
    
    
    def __repr__(self):
        """Returns the embedded activity's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        repr_parts.append(', channel_id = ')
        repr_parts.append(repr(self.channel_id))
        
        repr_parts.append(', user_ids = ')
        repr_parts.append(repr(self.user_ids))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two embedded activities are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        # activity.application_id
        if self.activity.application_id != other.activity.application_id:
            return False
        
        # channel_id
        if self.channel_id != other.channel_id:
            return False
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the hash value of the embedded activity state."""
        hash_value = 0
        
        # activity
        hash_value ^= self.activity.application_id
        
        # channel_id
        hash_value ^= self.channel_id
        
        # guild_id
        hash_value ^= self.guild_id
        
        return hash_value
    
    
    @property
    def application_id(self):
        """
        Returns the embedded activity's application's identifier.
        
        Returns
        -------
        application_id : `int`
        """
        return self.activity.application_id
    
    
    @property
    def key(self):
        """
        Returns the key used to store the embedded activity in cache.
        
        Returns
        -------
        key : ``EmbeddedActivityStateKey``
        """
        return EmbeddedActivityStateKey(self.guild_id, self.channel_id, self.application_id)
    
    
    @property
    def guild(self):
        """
        Returns the guild where the embedded activity is. The guild must be cached.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        return GUILDS.get(self.guild_id, None)
    
    
    @property
    def channel(self):
        """
        Returns the channel where the embedded activity is. The channel must be cached.
        
        Returns
        -------
        channel : ``Channel``
        """
        return create_partial_channel_from_id(self.channel_id, ChannelType.unknown, self.guild_id)
    
    
    @property
    def users(self):
        """
        Returns the users of the embedded activity.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        """
        return [create_partial_user_from_id(user_id) for user_id in self.user_ids]
    
    
    def copy(self):
        """
        Copies the embedded activity state.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.activity = self.activity.copy()
        new.channel_id = self.channel_id
        new.guild_id = self.guild_id
        new.user_ids = self.user_ids.copy()
        return new
    
    
    def copy_with(self, *, activity = ..., channel_id = ..., guild_id = ..., user_ids = ...):
        """
        Copies the embedded activity state with the given fields.
        
        Parameters
        ----------
        activity : ``Activity``, Optional (Keyword only)
            The embedded activity.
        channel_id : `int`, ``Channel``, Optional (Keyword only)
            The respective channel or its identifier.
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The respective guild or its identifier.
        user_ids : `None`, `iterable` of (`int`, ``ClientUserBase``), Optional (Keyword only)
            The joined users' identifiers.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's value is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # activity
        if activity is ...:
            activity = self.activity.copy()
        else:
            activity = validate_activity(activity)
        
        # channel_id
        if channel_id is ...:
            channel_id = self.channel_id
        else:
            channel_id = validate_channel_id(channel_id)
        
        # guild_id
        if guild_id is ...:
            guild_id = self.guild_id
        else:
            guild_id = validate_guild_id(guild_id)
        
        # user_ids
        if user_ids is ...:
            user_ids = self.user_ids.copy()
        else:
            user_ids = validate_user_ids(user_ids)
        
        new = object.__new__(type(self))
        new.activity = activity
        new.channel_id = channel_id
        new.guild_id = guild_id
        new.user_ids = user_ids
        return new



__all__ = ('ApplicationCommandCountUpdate',  'VoiceServerUpdateEvent', 'WebhookUpdateEvent',)

from scarletio import copy_docs

from ..application_command import ApplicationCommandTargetType
from ..bases import EventBase
from ..channel import ChannelType, create_partial_channel_from_id
from ..core import GUILDS


class VoiceServerUpdateEvent(EventBase):
    """
    Represents a `VOICE_SERVER_UPDATE` event.
    
    Attributes
    ----------
    endpoint : `None`, `str`
        The voice server's host.
    guild_id : `int`
        The respective guild's identifier.
    token : `None`, `str`
        Voice connection token.
    
    Examples
    --------
    The event can be unpacked like:
    
    ```py
    guild_id, endpoint, token = event
    ```
    """
    __slots__ = ('endpoint', 'guild_id', 'token')
    
    def __new__(cls, data):
        """
        Creates a new voice server update event from the given voice server update data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Voice server update data.
        """
        guild_id = data.get('guild_id', None)
        if guild_id is None:
            guild_id = 0
        else:
            guild_id = int(guild_id)
        
        self = object.__new__(cls)
        self.guild_id = guild_id
        self.endpoint = data.get('endpoint', None)
        self.token = data.get('token', None)
        
        return self
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__,]
        
        repr_parts.append(' guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        endpoint = self.endpoint
        if (endpoint is not None):
            repr_parts.append(', endpoint = ')
            repr_parts.append(repr(endpoint))
        
        token = self.token
        if (token is not None):
            repr_parts.append(', token = ')
            repr_parts.append(repr(self.token))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 3
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.guild_id
        yield self.endpoint
        yield self.token
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.endpoint != other.endpoint:
            return False
        
        if self.guild_id != other.guild_id:
            return False
        
        if self.token != other.token:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # endpoint
        endpoint = self.endpoint
        if (endpoint is not None):
            hash_value ^= hash(endpoint)
        
        # guild_id
        hash_value ^= self.guild_id
        
        # token
        token = self.token
        if (token is not None):
            hash_value ^= hash(token)
        
        return hash_value
    
    
    @property
    def guild(self):
        """
        Returns the event's guild from cache.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)


class WebhookUpdateEvent(EventBase):
    """
    Represents a `WEBHOOKS_UPDATE` event.
    
    Attributes
    ----------
    channel_id : `int`
        The respective channel's identifier.
    guild_id : `int`
        The respective guild's identifier.
    
    Examples
    --------
    The event can be unpacked like:
    
    ```py
    guild_id, channel_id = event
    ```
    """
    __slots__ = ('channel_id', 'guild_id')
    
    def __new__(cls, data):
        """
        Creates a new webhook update event from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Webhook update data.
        """
        channel_id = int(data['channel_id'])
        
        guild_id = data.get('guild_id', None)
        # non guild webhook check ??
        if guild_id is None:
            guild_id = 0
        else:
            guild_id = int(guild_id)
        
        self = object.__new__(cls)
        self.channel_id = channel_id
        self.guild_id = guild_id
        
        return self
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__,]
        
        repr_parts.append(' guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        repr_parts.append(', channel_id = ')
        repr_parts.append(repr(self.channel_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 2
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.guild_id
        yield self.channel_id
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # channel_id
        if self.channel_id != other.channel_id:
            return False
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # channel_id
        hash_value ^= self.channel_id
        
        # guild_id
        hash_value ^= self.guild_id
        
        return hash_value
    
    
    @property
    def channel(self):
        """
        Returns the event's channel from cache.
        
        Returns
        -------
        channel : ``Channel``
        """
        return create_partial_channel_from_id(self.channel_id, ChannelType.unknown, self.guild_id)
    
    
    @property
    def guild(self):
        """
        Returns the event's guild from cache.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)


class ApplicationCommandCountUpdate(EventBase):
    """
    Represents a `GUILD_APPLICATION_COMMAND_INDEX_UPDATE` event.
    
    Attributes
    ----------
    counts : `dict` of (``ApplicationCommandTargetType``, `int`) items
        The count of the application commands by target type.
    guild_id : `int`
        The respective guild's identifier.
    
    Examples
    --------
    The event can be unpacked like:
    
    ```py
    guild_id, channel_id = event
    ```
    """
    __slots__ = ('counts', 'guild_id')
    
    def __new__(cls, data):
        """
        Creates a new application command count update event from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Application command count update data.
        """
        raw_counts = data['application_command_counts']
        
        counts = {
            ApplicationCommandTargetType.get(int(application_command_target_type_string)): count
            for application_command_target_type_string, count in raw_counts.items()
        }
        
        guild_id = data.get('guild_id', None)
        # non guild check ??
        if guild_id is None:
            guild_id = 0
        else:
            guild_id = int(guild_id)
        
        self = object.__new__(cls)
        self.counts = counts
        self.guild_id = guild_id
        
        return self
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__,]
        
        repr_parts.append(' guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        repr_parts.append(', counts = {')
        
        items = sorted(self.counts.items())
        item_count = len(items)
        if item_count:
            index = 0
            while True:
                application_command_target_type, count = items[index]
                repr_parts.append(application_command_target_type.name)
                repr_parts.append(': ')
                repr_parts.append(str(count))
                
                index += 1
                if index == item_count:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append('}')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 2
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.guild_id
        yield self.counts
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # counts
        if self.counts != other.counts:
            return False
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # counts
        for application_command_target_type, count in self.counts.items():
            hash_value ^= application_command_target_type * count * count
        
        # guild_id
        hash_value ^= self.guild_id
        
        return hash_value
    
    
    @property
    def guild(self):
        """
        Returns the event's guild from cache.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)



__all__ = ('create_partial_sticker_data', 'create_partial_sticker_from_partial_data')


from .fields import put_format_into, put_id_into, put_name_into, parse_id, parse_format, parse_name
from .sticker import Sticker
from ...core import STICKERS


def create_partial_sticker_data(sticker):
    """
    Creates partial sticker data for the given sticker.
    
    Parameters
    ----------
    sticker : ``Sticker``
        The sticker to create the partial data from.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    data = {}
    put_format_into(sticker.format, data, True)
    put_id_into(sticker.id, data, True)
    put_name_into(sticker.name, data, True)
    return data


def create_partial_sticker_from_partial_data(data):
    """
    Creates a sticker from the given partial sticker data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Sticker data.
    
    Returns
    -------
    sticker : `instance<cls>`
    """
    sticker_id = parse_id(data)
    
    try:
        self = STICKERS[sticker_id]
    except KeyError:
        self = Sticker._create_empty(sticker_id)
        STICKERS[sticker_id] = self
    else:
        if not self.partial:
            return self
    
    self.format = parse_format(data)
    self.name = parse_name(data)
    
    return self



__all__ = ()

from scarletio import include

from ...channel import ChannelType
from ...field_parsers import (
    bool_parser_factory, int_parser_factory, negated_bool_parser_factory, nullable_object_array_parser_factory,
    nullable_string_parser_factory, preinstanced_array_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, int_optional_putter_factory, negated_bool_optional_putter_factory,
    nullable_entity_array_putter_factory, nullable_string_optional_putter_factory, preinstanced_array_putter_factory,
    preinstanced_putter_factory, url_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, int_conditional_validator_factory, nullable_object_array_validator_factory,
    nullable_string_validator_factory, preinstanced_array_validator_factory, preinstanced_validator_factory,
    url_optional_validator_factory
)

from ..string_select_option import StringSelectOption

from .constants import (
    BUTTON_STYLE_DEFAULT, LABEL_LENGTH_MAX, MAX_LENGTH_DEFAULT, MAX_LENGTH_MAX, MAX_LENGTH_MIN, MAX_VALUES_DEFAULT,
    MAX_VALUES_MAX, MAX_VALUES_MIN, MIN_LENGTH_DEFAULT, MIN_LENGTH_MAX, MIN_LENGTH_MIN, MIN_VALUES_DEFAULT,
    MIN_VALUES_MAX, MIN_VALUES_MIN, PLACEHOLDER_LENGTH_MAX, TEXT_INPUT_STYLE_DEFAULT, VALUE_LENGTH_MAX
)
from .preinstanced import ButtonStyle, TextInputStyle


Component = include('Component')

# button_style

parse_button_style = preinstanced_parser_factory('style', ButtonStyle, BUTTON_STYLE_DEFAULT)
put_button_style_into = preinstanced_putter_factory('style')
validate_button_style = preinstanced_validator_factory('button_style', ButtonStyle)

# channel_types

parse_channel_types = preinstanced_array_parser_factory('channel_types', ChannelType)
put_channel_types_into = preinstanced_array_putter_factory('channel_types')
validate_channel_types = preinstanced_array_validator_factory('channel_types', ChannelType)

# enabled

parse_enabled = negated_bool_parser_factory('disabled', True)
put_enabled_into = negated_bool_optional_putter_factory('disabled', True)
validate_enabled = bool_validator_factory('enabled', True)

# label

parse_label = nullable_string_parser_factory('label')
put_label_into = nullable_string_optional_putter_factory('label')
validate_label = nullable_string_validator_factory('label', 0, LABEL_LENGTH_MAX)

# max_length

parse_max_length = int_parser_factory('max_length', MAX_LENGTH_DEFAULT)
put_max_length_into = int_optional_putter_factory('max_length', MAX_LENGTH_DEFAULT)
validate_max_length = int_conditional_validator_factory(
    'max_length',
    MAX_LENGTH_MIN,
    (lambda user_limit: user_limit >= MAX_LENGTH_MIN and user_limit <= MAX_LENGTH_MAX),
    f'>= {MAX_LENGTH_MIN} and <= {MAX_LENGTH_MAX},'
)

# max_values

parse_max_values = int_parser_factory('max_values', MAX_VALUES_DEFAULT)
put_max_values_into = int_optional_putter_factory('max_values', MAX_VALUES_DEFAULT)
validate_max_values = int_conditional_validator_factory(
    'max_values',
    MAX_VALUES_MIN,
    (lambda user_limit: user_limit >= MAX_VALUES_MIN and user_limit <= MAX_VALUES_MAX),
    f'>= {MAX_VALUES_MIN} and <= {MAX_VALUES_MAX},'
)

# min_length

parse_min_length = int_parser_factory('min_length', MIN_LENGTH_DEFAULT)
put_min_length_into = int_optional_putter_factory('min_length', MIN_LENGTH_DEFAULT)
validate_min_length = int_conditional_validator_factory(
    'min_length',
    MIN_LENGTH_MIN,
    (lambda user_limit: user_limit >= MIN_LENGTH_MIN and user_limit <= MIN_LENGTH_MAX),
    f'>= {MIN_LENGTH_MIN} and <= {MIN_LENGTH_MAX},'
)

# min_values

parse_min_values = int_parser_factory('min_values', MIN_VALUES_DEFAULT)
put_min_values_into = int_optional_putter_factory('min_values', MIN_VALUES_DEFAULT)
validate_min_values = int_conditional_validator_factory(
    'min_values',
    MIN_VALUES_MIN,
    (lambda user_limit: user_limit >= MIN_VALUES_MIN and user_limit <= MIN_VALUES_MAX),
    f'>= {MIN_VALUES_MIN} and <= {MIN_VALUES_MAX},'
)

# options

parse_options = nullable_object_array_parser_factory('options', StringSelectOption)
put_options_into = nullable_entity_array_putter_factory('options', StringSelectOption)
validate_options = nullable_object_array_validator_factory('options', StringSelectOption)

# placeholder

parse_placeholder = nullable_string_parser_factory('placeholder')
put_placeholder_into = nullable_string_optional_putter_factory('placeholder')
validate_placeholder = nullable_string_validator_factory('placeholder', 0, PLACEHOLDER_LENGTH_MAX)

# required

parse_required = bool_parser_factory('required', True)
put_required_into = bool_optional_putter_factory('required', True)
validate_required = bool_validator_factory('required', True)

# text_input_style

parse_text_input_style = preinstanced_parser_factory('style', TextInputStyle, TEXT_INPUT_STYLE_DEFAULT)
put_text_input_style_into = preinstanced_putter_factory('style')
validate_text_input_style = preinstanced_validator_factory('text_input_style', TextInputStyle)

# url

parse_url = nullable_string_parser_factory('url')
put_url_into = url_optional_putter_factory('url')
validate_url = url_optional_validator_factory('url')

# value

parse_value = nullable_string_parser_factory('value')
put_value_into = nullable_string_optional_putter_factory('value')
validate_value = nullable_string_validator_factory('value', 0, VALUE_LENGTH_MAX)
